import pytest
import numpy as np
import numba
from neutralocean.fzero import brent_guess

tol = 1e-6

# Test with univariate function
@numba.njit
def univar(x):
    return x - 2.5


@numba.njit
def univar3(x):
    return (x - 2.5) ** 3


@pytest.mark.parametrize("func", [univar, univar3])
def test_root(func):
    root = brent_guess(func, 1.0, 0.0, 6.0, tol)
    assert abs(root - 2.5) < tol


# Test when one end of the search range is a root
def test_ub_eq_root():
    root = brent_guess(univar, 1.0, 0.0, 2.5, tol)
    assert abs(root - 2.5) < tol


def test_lb_eq_root():
    root = brent_guess(univar, 4.0, 2.5, 6.0, tol)
    assert abs(root - 2.5) < tol


# Test function with extra parameters
@numba.njit
def univar_args(x, exp):
    return (x - 2.5) ** exp * (x + 2.5) ** exp


@pytest.mark.parametrize("args", [(1,), (3,)])
def test_root_args(args):
    root = brent_guess(univar_args, 1.0, 0.0, 6.0, tol, args)
    assert abs(root - 2.5) < tol


# Test function which does not change sign at its roots, i.e. a quadratic
def test_root_singular():
    root = brent_guess(univar_args, 1.0, 0.0, 6.0, tol, (2,))
    assert np.isnan(root)


# Test function which does not change sign at its roots, but which is zero at
# one end of the search range
def test_ub_eq_root_singular():
    root = brent_guess(univar_args, 1.0, 0.0, 2.5, tol, (2,))
    assert abs(root - 2.5) < tol


def test_lb_eq_root_singular():
    root = brent_guess(univar_args, 4.0, 2.5, 6.0, tol, (2,))
    assert abs(root - 2.5) < tol



import numpy as np
import pytest

from neutralocean.ppinterp import make_pp, ppval, valid_range_1
from scipy.interpolate import UnivariateSpline, PchipInterpolator

NX = 6  # number of 1D interpolation problems
NY = 4
K = 10  # number of grid points in each interpolation problem

X1 = np.linspace(0, 10, K) ** 1.2  # Monotonic but non-uniform independent data
X = np.tile(X1, (NX, NY, 1))
X[1, :, 7:K] = np.nan  # Test NaN's at end
X[2, :, 1:] = np.nan  # Test almost all NaN's, except X[0]
X[3, :, 0:2] = np.nan  # Test NaN's at start, ...
X[3, :, 7:K] = np.nan  # ... and end
X[4, :, 0:5] = np.nan  # Test 2 valid points, ...
X[4, :, 7:K] = np.nan  # ... in the middle
X[5, :, :] = np.nan  # Test all NaN

# Build dependent data
Y = np.empty((NX, NY, K), dtype=np.float64)
Y[:, 0:2] = np.ones(K) * (X1 / X1[-1]) ** 2  # steadily increasing
Y[:, 2] = np.sin(X1 / X1[-1] * 2 * np.pi)  # smooth wave

Y[:, 3] = Y[:, 1] * np.cos(X1 / X1[-1] * 10 * np.pi)  # crazy wave ...
Y[:, 3, -3:] = np.nan  # ... with NaN's at end

# Reshape all casts into a 1D array of 1D casts
N = NX * NY
X, Y = [np.reshape(Z, (N, K)) for Z in (X, Y)]

# Interpolate between each knot
x_midpts = X1[0:-1] + np.diff(X1) / 2

# Interpolate to each knot point and each midpoint
x_targets = np.sort(np.concatenate((X1, x_midpts)))

Y1 = Y[0, :]

# Note re: lhs vs rhs:
# Our code won't agree with SciPy when evaluating a function that is piecewise
# discontinuous (e.g. first derivative of a linear interpolant or second
# derivative of a PHCIP) at the knots, because we evaluate using the "left"
# side whereas SciPy evaluates using the "right" side.
@pytest.mark.parametrize(
    "interp,num_deriv,x",
    [
        ("linear", 0, x_targets),
        ("linear", 1, x_midpts),  # see "Note re: lhs vs rhs"
        ("pchip", 0, x_targets),
        ("pchip", 1, x_targets),
        ("pchip", 2, x_midpts),  # see "Note re: lhs vs rhs"
    ],
)
def test_interp(interp, num_deriv, x):

    # Interpolate with SciPy
    y = np.full((N, x.size), np.nan, dtype=float)
    for i in range(N):
        k, K = valid_range_1(X[i] + Y[i])
        try:
            if interp == "linear":
                fn = UnivariateSpline(
                    X[i, k:K], Y[i, k:K], k=1, s=0, ext="raise"
                )
            elif interp == "pchip":
                fn = PchipInterpolator(X[i, k:K], Y[i, k:K], extrapolate=False)

            fn = fn.derivative(num_deriv)
            for j in range(x.size):
                try:
                    y[i, j] = fn(x[j])
                except:
                    # extrapolation was needed (only for UnivariateSpline)
                    pass  # leave as nan
        except:
            # Building the interpolant failed.
            # Not enough valid data points (e.g. X has 1 non-nan value)
            pass  # leave as nan

    # Interpolate with our methods: first, on the fly using ppinterp
    interp_fn = make_pp(interp, out="interp", kind="u")
    y1 = np.empty_like(y)
    for j in range(x.size):
        y1[:, j] = interp_fn(x[j], X, Y, num_deriv)

    # Verify our method is close to SciPy. Won't be bitwise identical.
    assert np.allclose(y, y1, equal_nan=True)

    # Interpolate with our methods:
    # second, with piecewise polynomial coefficients, using ppinterp
    ppc_fn = make_pp(interp, out="coeffs", kind="u")
    Yppc = ppc_fn(X, Y)
    y2 = np.empty_like(y)
    for j in range(x.size):
        y2[:, j] = ppval(x[j], X, Yppc, num_deriv)

    # Our interpolation methods, whether on-the-fly (y1) or building the PPCs
    # for the entire interpolant (y2), should be bitwise identical.
    assert np.array_equal(y1, y2, equal_nan=True)



from neutralocean.lib import aggsum


def divergence(vector, edges, distperp=1.0, area=1.0):
    """
    Calculate the divergence of a vector field defined on the edges of a graph

    `vector[i]` is positive when the vector field points from node `a[i]` to
    node `b[i]`, where `(a,b) = edges`.
    `distperp[i]` is the distance of the interface between nodes `a[i]` and `b[i]`.
    `area[i]` is the area of node `i`.
    
    Note: This function's API is likely to change, to just take in a `grid` dict.
    """

    a, b = edges
    N = max(max(a), max(b)) + 1  # number of nodes
    e = vector * distperp
    divg = (aggsum(e, a, N) - aggsum(e, b, N)) / area
    return divg



import numpy as np
import xarray as xr
import xgcm


def build_grid(
    n, face_connections, dims, xsh, ysh, dxC=1.0, dyC=1.0, dyG=1.0, dxG=1.0
):
    """
    Make a list of edges between adjacent nodes in a graph consisting of
    rectilinear tiles, with adjacency between tiles specified as for xgcm;
    also generate lists of distances associated with these edges.

    Here, the grid is thought of as a graph, where each node represents a
    water column, and the graph has an edge between two nodes when the
    corresponding water columns are adjacent in physical space.


    Parameters
    ----------
    n : int
        Number of grid points on each side of a square tile.
        (Note xgcm only handles square tiles.)

    face_connections : dict
        xgcm-style specification for how tiles (faces) are connected.
        See https://xgcm.readthedocs.io/en/latest/grid_topology.html

    dims : tuple of str
        A tuple of length 3, with one entry being the same as the only key in
        the `face_connections` dict, and the other two being 'i' and 'j'.

        This names, in order, the dimensions of a DataArray whose data lives
        on each (tracer) grid point in the horizontal, and has no vertical
        dimension.

        If `D` is a DataArray giving the Sea Surface Temperature, for example,
        then this argument should be `D.dims`.

        Example: in ECCOv4r4, `dims == ('tile', 'j', 'i')`.

    xsh : str
        The direction of shifting in the 'i' dimension.  Can be either "left"
        or "right".

        If "left", then `dxC` is the distance between the local grid point (i, j)
        and the previous grid point in the 'i' dimension, at (i-1, j).

        If "right", then `dxC` is the distance between the local grid point (i, j)
        and the next grid point in the 'i' dimension, at (i+1, j).

    ysh : str
        The direction of shifting in the 'j' dimension.  Can be either "left"
        or "right".

        If "left", then `dyC` is the distance between the local grid point (i, j)
        and the previous grid point in the 'j' dimension, at (i, j+1).

        If "right", then `dyC` is the distance between the local grid point (i, j)
        and the next grid point in the 'j' dimension, at (i, j+1).

    dxC : float or ndarray, Default 1.0
        Distance between adjacent grid points in the 'i' dimension.

    dyC : float or ndarray, Default 1.0
        Distance between adjacent grid points in the 'j' dimension

    dyG : float or ndarray, Default 1.0
        Distance (in the 'j' dimension) of the face between grid points that
        are adjacent in the 'i' dimension.
        Lives at the same location as `dxC`.

    dxG : float or ndarray, Default 1.0
        Distance (in the 'i' dimension) of the face between grid points that
        are adjacent in the 'j' dimension.
        Lives at the same location as `dyC`.


    Returns
    -------
    grid : dict
        Containing the following:

        edges : tuple of length 2
            Each element is an array of int of length E, where E is the number of
            edges in the grid's graph, i.e. the number of pairs of adjacent water
            columns (including land) in the grid.
            If `edges = (a, b)`, the nodes (water columns) whose linear indices are
            `a[i]` and `b[i]` are adjacent.

        dist : 1d array
            Horizontal distance between adjacent water columns (nodes).
            `dist[i]` is the distance between nodes whose linear indices are
            `edges[0][i]` and `edges[1][i]`.

        distperp : 1d array
            Horizontal distance of the face between adjacent water columns (nodes).
            `distperp[i]` is the distance of the interface between nodes whose
            linear indices are `edges[0][i]` and `edges[1][i]`.

    """

    # Process inputs
    assert xsh[0] in ("l", "r"), "Expected xsh to be either 'left' or 'right'"
    assert ysh[0] in ("l", "r"), "Expected ysh to be either 'left' or 'right'"

    # If given DataArrays, check the dims are correct then convert to numpy arrays.
    if all(hasattr(x, "dims") for x in (dxC, dyC, dyG, dxG)):
        assert (
            dxC.dims == dyG.dims
        ), "Expected dxC and dyG to have the same coordinates"
        assert (
            dyC.dims == dxG.dims
        ), "Expected dyC and dxG to have the same coordinates"
    dxC, dyC, dyG, dxG = (
        x.values if hasattr(x, "values") else x for x in (dxC, dyC, dyG, dxG)
    )

    im1, jm1 = _build_im1_jm1(n, face_connections, dims, xsh, ysh)
    N = im1.size

    grid = dict()

    # Build list of pairs of adjacent nodes.  The first N entries are [m,n]
    # where m is the local node and n is its neighbour in the j dimension.
    # The second N entries are [m,n] where m is the local node and n is its
    # neighbour in the i dimension.
    a = np.tile(np.arange(N), 2)
    b = np.empty(N * 2, dtype=int)  # prealloc space
    b[:N] = im1.values.reshape(-1)
    b[N:] = jm1.values.reshape(-1)

    # Build list of distances between adjacent nodes, in the same order as the
    # pairs of nodes built by edges.
    dist = np.empty(N * 2, dtype=float)
    dist[:N] = dyC.reshape(-1)
    dist[N:] = dxC.reshape(-1)

    # Build list of distances of the faces between adjacent nodes, in the same
    # order as the pairs of nodes built by edges.
    distperp = np.empty(N * 2, dtype=float)
    distperp[:N] = dxG.reshape(-1)  # dxG lives at same place as dyC
    distperp[N:] = dyG.reshape(-1)  # dyG lives at same place as dxC

    # Trim out invalid edges, i.e. those for which one of the two nodes was
    # filled by `apply_as_grid_ufunc` to have a value of -1.
    good = b >= 0
    a, b, dist, distperp = (x[good] for x in (a, b, dist, distperp))

    grid["edges"] = (a, b)
    grid["dist"] = dist
    grid["distperp"] = distperp
    return grid


def edgedata_to_maps(edgedata, n, face_connections, dims, xsh, ysh):
    """
    Convert 1D array of data living on edges into two nD arrays, one for each
    spatial dimension, for a rectilinear grid

    Parameters
    ----------
    edgedata : array
        1D array of data that lives on edges of the grid's graph, given in the
        same order as the edges cosntructed from `build_grid`.
        If `a, b, ... = build_grid(...)`, then `edgedata[i]` is the
        data that lives at the interface between the grid cells indexed by
        `a[i]` and `b[i]`.

    n, face_connections, dims, xsh, ysh :
        See `build_grid`

    Returns
    -------
    Fi, Fj : ndarray
        `Fi` contains the values from `edgedata` that correspond to data living
        between grid points in the first dimension, and similarly for `Fj` but
        in the second dimension.
    """

    im1, jm1 = _build_im1_jm1(n, face_connections, dims, xsh, ysh)
    N = im1.size

    Fi, Fj = (np.full(N, np.nan) for x in (0, 1))

    # Build linear indices for grid cells (i,j) where (i-1,j) is valid
    idx_goodim1 = np.flatnonzero(im1 >= 0)
    Ni = len(idx_goodim1)

    Fi[idx_goodim1] = edgedata[:Ni]
    Fj[np.flatnonzero(jm1 >= 0)] = edgedata[Ni:]

    Fi, Fj = (np.reshape(x, im1.shape) for x in (Fi, Fj))
    return Fi, Fj


def _build_im1_jm1(n, face_connections, dims, xsh, ysh):

    # Access first (and only) key / value of face_connections
    tile = next(
        iter(face_connections.keys())
    )  # name of face_connections only key
    nf = len(next(iter(face_connections.values())))  #  number of faces

    # Get index of tile in dims
    try:
        t = dims.index(tile)
    except:
        raise ValueError(
            f"Expected to find '{tile}' (the key from `face_connections`) in `dims` == {dims}."
        )

    # Ensure dims also has 'i' and 'j':
    if not (dims.__contains__("i") and dims.__contains__("j")):
        raise ValueError(f"Expected to find 'i' and 'j' in `dims` == {dims}.")

    N = nf * n * n  # number of Nodes (water columns)

    # Shape of water columns (e.g. shape of ndarray storing sea surface temperature data)
    shape = [n, n, n]
    shape[t] = nf

    # Build a simple xgcm grid
    coords = {"i": np.arange(n), "j": np.arange(n), tile: np.arange(nf)}
    grid = xgcm.Grid(
        xr.Dataset(None, coords),
        periodic=False,
        face_connections=face_connections,
        coords={
            "X": {"center": "i", "left": "i", "right": "i"},
            "Y": {"center": "j", "left": "j", "right": "j"},
        },
    )

    # Make linear indices for the whole domain: a DataArray of 0, 1, ... N-1.
    idx = xr.DataArray(np.arange(N).reshape(shape), dims=dims, coords=coords)

    # Make linear indices to the neighbour in the i//x dimension.
    # That is, idx[t,j,i-1] == im1[t,j,i] for all 0 <= t < nf, 0 <= j < n, 0 < i < n
    # and corresponding treatment across tiles when i == 0.
    # Similarly, idx[t,j-1,i] == jm1[t,j,i] for all 0 <= t < nf, 0 < j < n, 0 <= i < n
    # and corresponding treatment across tiles when j == 0.
    # To achieve this, use XGCM grid Ufunc to pad the `idx` data as needed,
    # then subset forwards or backwards.
    if xsh[0] == "l":  # left
        # # Old code using depreciated XGCM function below
        # im1 = grid.axes["X"]._neighbor_binary_func(
        #     idx, lambda a, b: a, to="left", boundary="fill", fill_value=-1
        # )
        im1 = grid.apply_as_grid_ufunc(
            lambda a: a[:, :, 0:-1],
            idx,
            axis=[["X"]],
            signature="(ax1:center)->(ax1:center)",
            boundary_width={"ax1": (1, 0)},
            boundary="fill",
            fill_value=-1,
        )

    else:  # right
        # im1 = grid.axes["X"]._neighbor_binary_func(
        #     idx, lambda a, b: b, to="right", boundary="fill", fill_value=-1
        # )
        im1 = grid.apply_as_grid_ufunc(
            lambda a: a[:, :, 1:],
            idx,
            axis=[["X"]],
            signature="(ax1:center)->(ax1:center)",
            boundary_width={"ax1": (0, 1)},
            boundary="fill",
            fill_value=-1,
        )

    # Make linear indices to the neighbour in the j//y dimension.
    # Why do the following lambda functions subset along the last dimension,
    # when j is the second (middle) dimension?  See:
    # https://xgcm.readthedocs.io/en/latest/grid_ufuncs.html
    # "XGCM assumes the function acts along the last axis of the numpy array"
    if ysh[0] == "l":  # left
        # jm1 = grid.axes["Y"]._neighbor_binary_func(
        #     idx, lambda a, b: a, to="left", boundary="fill", fill_value=-1
        # )
        jm1 = grid.apply_as_grid_ufunc(
            lambda a: a[:, :, 0:-1],
            idx,
            axis=[["Y"]],
            signature="(ax1:center)->(ax1:center)",
            boundary_width={"ax1": (1, 0)},
            boundary="fill",
            fill_value=-1,
        )

        # Must ensure same ordering of dimensions as idx.  Currently, this
        # one gets its dimensions reordered by the grid ufunc, probably
        # because it reorders things to act on the last dimension...?
        jm1 = jm1.transpose(*idx.dims)

    else:  # right
        # jm1 = grid.axes["Y"]._neighbor_binary_func(
        #     idx, lambda a, b: b, to="right", boundary="fill", fill_value=-1
        # )
        jm1 = grid.apply_as_grid_ufunc(
            lambda a: a[:, :, 1:],
            idx,
            axis=[["Y"]],
            signature="(ax1:center)->(ax1:center)",
            boundary_width={"ax1": (0, 1)},
            boundary="fill",
            fill_value=-1,
        )

    return im1, jm1



"""
Functions for finding the zero of a univariate function.
"""

import numpy as np
import numba as nb


eps = np.finfo(np.float64).eps


@nb.njit
def brent_guess(f, x, A, B, t, args=()):
    """
    Find a zero of a function within a given range, starting from a guess

    Parameters
    ----------
    f : function
        Continuous function of a single variable.
    x : float
        initial guess for a root
    A, B : float
        Range within which to search, satisfying `A < B`
    t : float
        Tolerance for convergence.
    args : tuple
        Additional arguments, beyond the optimization argument, to be passed to `f`.
        Pass `()` when `f` is univariate.

    Returns
    -------
    float
        Value of `x` where `f(x) ~ 0`.

    """

    a, b = guess_to_bounds(f, x, A, B, args)
    return brent(f, a, b, t, args)


@nb.njit
def brent(f, a, b, t, args=()):
    """
    Find a zero of a univariate function within a given range

    This is a bracketed root-finding method, so `f(a)` and `f(b)` must differ in
    sign. If they do, a root is guaranteed to be found.

    Parameters
    ----------
    f : function
        Continuous function of a single variable.
    a, b : float
        Range within which to search, satisfying `a < b` and ideally `f(a) * f(b) <= 0`
    t : float
        Tolerance for convergence.
    args : tuple
        Additional arguments, beyond the optimization argument, to be passed to `f`.
        Pass `()` when `f` is univariate.

    Returns
    -------
    float
        Value of `x` where `f(x) ~ 0`.

    Notes
    -----
    `f` should be a `@numba.njit`'ed function (when this function is `njit`'ed).
    """

    # Protection against bad input search range
    if np.isnan(a) or np.isnan(b) or a > b:
        return np.nan

    fa = f(a, *args)
    fb = f(b, *args)

    # Protection against input range that doesn't have a sign change
    if fa * fb > 0:  # DEV note: check if this should be fa * fb >= 0
        return np.nan

    c = a
    fc = fa
    e = b - a
    d = e

    while True:
        if abs(fc) < abs(fb):
            a = b
            b = c
            c = a
            fa = fb
            fb = fc
            fc = fa

        tol = 2.0 * eps * abs(b) + t
        m = 0.5 * (c - b)

        if abs(m) <= tol or fb == 0.0:
            break

        if abs(e) < tol or abs(fa) <= abs(fb):
            e = m
            d = e
        else:
            s = fb / fa
            if a == c:
                p = 2.0 * m * s
                q = 1.0 - s
            else:
                q = fa / fc
                r = fb / fc
                p = s * (2.0 * m * q * (q - r) - (b - a) * (r - 1.0))
                q = (q - 1.0) * (r - 1.0) * (s - 1.0)

            if 0.0 < p:
                q = -q
            else:
                p = -p

            s = e
            e = d

            if 2.0 * p < 3.0 * m * q - abs(tol * q) and p < abs(0.5 * s * q):
                d = p / q
            else:
                e = m
                d = e

        a = b
        fa = fb

        if tol < abs(d):
            b += d
        elif 0.0 < m:
            b += tol
        else:
            b -= tol

        fb = f(b, *args)

        if (0.0 < fb and 0.0 < fc) or (fb <= 0.0 and fc <= 0.0):
            c = a
            fc = fa
            e = b - a
            d = e

    return b


@nb.njit
def guess_to_bounds(f, x, A, B, args=()):
    """
    Search for a range containing a sign change, expanding geometrically
    outwards from the initial guess.

    This is used as a first step in zero-finding, providing a small search
    range for the Brent algorithm.

    Parameters
    ----------
    f : function
        Continuous function of a single variable
    x : float
        Central point for starting the search
    A, B : float
        Lower and upper bounds, containing `x`, within which to search for a zero.
    args : tuple
        Additional arguments beyond the optimization argument.
        Pass `()` when `f` is univariate.

    Returns
    -------
    a, b : float
        Lower and upper bounds within which `f(x)` changes sign.
    """

    nan = np.nan

    # Check value of f at bounds
    fa = f(A, *args)
    if fa == 0.0:
        return (A, A)

    fb = f(B, *args)
    if fb == 0.0:
        return (B, B)

    x = min(max(x, A), B)

    # initial distance to expand outward from x, in positive and negative directions
    dxp = (B - x) / 50
    dxm = (x - A) / 50

    # Set a = x, except when x is so close to A that machine roundoff makes dxm identically 0
    # which would lead to an infinite loop below.  In this case, set a = A.
    if dxm == 0:
        a = A
    else:
        a = x

    # Similarly, set b = x, except for machine precision problems.
    if dxp == 0:
        b = B
    else:
        b = x

    if a > A:
        fbpos = f(b, *args) > 0.0
    else:  # a == A
        if b == B:
            # So dxm == 0 and dxp == 0.  So A very nearly equals B, but could
            # have A != B due to machine precision problems
            fapos = f(a, *args) > 0.0
            fbpos = f(b, *args) > 0.0
        else:  # b < B
            fapos = f(a, *args) > 0.0

    while True:
        if a > A:
            # Move a left, and test for a sign change
            dxm *= 1.414213562373095
            a = max(x - dxm, A)
            fapos = f(a, *args) > 0.0
            if fapos != fbpos:  # fa and fb have different signs
                return (a, b)
        elif b == B:  # also a == A, so cannot expand anymore
            if fapos != fbpos:  # one last test for sign change
                return (a, b)
            else:  # no sign change found
                return (nan, nan)

        if b < B:
            # Move b right, and test for a sign change
            dxp *= 1.414213562373095
            b = min(x + dxp, B)
            fbpos = f(b, *args) > 0.0
            if fapos != fbpos:  # fa and fb have different signs
                return (a, b)
        elif a == A:  # also b == B, so cannot expand anymore
            if fapos != fbpos:  # one last test for sign change
                return (a, b)
            else:  # no sign change found
                return (nan, nan)



"""
Functions to evaluate piecewise polynomials, given their coefficients, in one
dimension.
"""


"""
Adapted from 'Piecewise Polynomial Calculus' available on the MATLAB
File Exchange at
https://mathworks.com/matlabcentral/fileexchange/73114-piecewise-polynomial-calculus
"""

import numpy as np
import numba as nb

from .lib import valid_range_1_two


@nb.njit
def pval(x, C, d=0):
    """
    Evaluate a (derivative of a) single polynomial.

    Parameters
    ----------
    x : float
        Evaluation site
    C : array of float
        Polynomial coefficients, starting with the highest order coefficient
        and ending with the coefficient of the constant term.
        The order of the polynomial is `len(C)`.
    d : int, Default 0
        Order of the derivative to evaluate. When `0`, just evaluate the
        polynomial.

    Returns
    -------
    y : float
        The polynomial (or its `d`'th derivative) evaluated at `x`.

    Example
    -------
    If `C == (C2, C1, C0)`, and `d == 0`, then
    `y == x * (x * C2 + C1) + C0`.
    """

    order = len(C)  # order of the polynomial

    # Evaluate polynomial, using nested multiplication.
    # E.g. the cubic case is:
    # y = x^3 * C[0] + x^2 * C[1] + x * C[2] + C[3]
    #   = x * (x * (x * C[0] + C[1]) + C[2]) + C[3]
    y = 0.0
    if d == 0:
        for j in range(0, order):
            y = y * x + C[j]
    else:
        # Evaluate polynomial derivative, using nested multiplication.
        # E.g. the second derivative of the cubic case is:
        # y = 6 * x * C[0] + 2 * C[1]
        y = 0.0
        for o in range(0, order - d):  # o for order
            p = 1.0  # Build integer multiplying the coefficient
            for a in range(order - o - d, order - o):
                p *= a
            # p = np.prod(np.arange(degree - o - d + 1, degree - o + 1))  # slow!
            y = y * x + C[o] * p

    return y


@nb.njit
def ppval_1(x, X, Yppc, d=0):
    """
    Evaluate a single Piecewise Polynomial (PP).

    Parameters
    ----------
    x : float
        Evaluation site
    X : ndarray, 1d
        Independent variable. Must be monotonically increasing.
    Yppc : ndarray, 2d
        Piecewise Polynomial Coefficients.  First dimension must match `len(X)`.
    d : int, Default 0
        Number of derivatives to take.  If 0, simply evaluate the PP.

    Returns
    -------
    y : float
        The value of the PP (or its `d`'th derivative) at `X = x`.

    Notes
    -----
    If `X` and `Y` have NaN's, only the first contiguous block of non-NaN data
    between both `X` and `Y` is used.

    """

    # Find k = index to first valid data site, and
    # K such that K - 1 = index to last valid data site in contiguous range
    #                     of valid data after index k.
    # Note, Yppc[:,-1] = Y, the original data
    k, K = valid_range_1_two(X, Yppc[:, -1])

    # K - 1 == k means there's only one valid data point. Can't interpolate that.
    if K - 1 <= k or np.isnan(x) or x < X[k] or X[K - 1] < x:
        return np.nan

    # i = searchsorted(X,x) is such that:
    #   i = 0                   if x <= X[0] or all(isnan(X))
    #   i = len(X)              if X[-1] < x or isnan(x)
    #   X[i-1] < x <= X[i]      otherwise
    #
    # So
    # i = searchsorted(X[k:K], x)  is such that:
    #   i = 0              if x <= X[k] or all(isnan(X[k:K]))
    #   i = K-k            if X[K-1] < x or isnan(x)
    #   X[k:K][i-1] < x <= X[k:K][i] otherwise
    #
    # Having guaranteed above that x is not nan, and X[k:K] is all non-NaN, and
    # X[k] <= x <= X[K-1], then
    #   i = (searchsorted(X[k:K],x) + k)
    # is such that
    #   k <= i <= K-1
    # is guaranteed, and
    #   x == X[k]           when i == k,
    #   X[i-1] < x <= X[i]  when k+1 <= i <= K-1.
    #
    # Next, merge i == k into the i == k+1 case:
    #   i = max(k+1, searchsorted(X[k:K], x) + k)
    # is such that
    #   k+1 <= i <= K-1
    # is guaranteed, and
    #   X[k]  <= x <= X[k+1]  when  i == k+1,
    #   X[i-1] < x <= X[i]    when  k+2 <= i <= K-1.
    #
    # Finally, subtract 1:
    #   i = max(k, searchsorted(X[k:K], x) + k - 1)
    # is such that
    #   k <= i <= K-2
    # is guaranteed, and
    #   X[k] <= x <= X[k+1]  when  i == k,
    #   X[i] <  x <= X[i+1]  when  k+1 <= i <= K-2.
    i = max(k, np.searchsorted(X[k:K], x) + k - 1)

    dx = x - X[i]  # >= 0
    return pval(dx, Yppc[i], d)


@nb.njit
def ppval_1_nonan(x, X, Yppc, d=0):
    """As `ppval_1` but knowing there are no NaN's in the inputs X and Yppc"""
    if np.isnan(x) or x < X[0] or X[-1] < x:
        return np.nan

    i = max(0, np.searchsorted(X, x) - 1)

    dx = x - X[i]  # >= 0
    return pval(dx, Yppc[i], d)


@nb.guvectorize(
    [(nb.f8, nb.f8[:], nb.f8[:, :], nb.i8, nb.f8[:])],
    "(),(n),(n,m),()->()",
)
def ppval(x, X, Yppc, d, y):
    """
    Evaluate a given Piecewise Polynomial

    Parameters
    ----------
    x : ndarray

        Sites at which to evaluate the independent variable.

        Must be broadcastable to the shape of `X` or `Y` less their final
        dimension.

    X : ndarray

        Independent variable. Must be broadcastable to the shape of `Y`.

        `X` must be monotonically increasing in its last dimension.
        That is, `X[*i,:]` must be monotonically increasing for any
        `i` tuple indexing all but the last dimension.

    Y : ndarray

        Dependent variable.  Must be broadcastable to the shape of `X`.

    Yppc : ndarray

        Piecewise Polynomial Coefficients for the interpolant.
        The shape of `Yppc` less its final dimension must be broadcastable to
        the shape of `X` and/or `Y`.

    Returns
    -------
    y : ndarray

        Value of the piecewise polynomial interpolant (or its `d`'th
        derivative) at `x`.

    Notes
    -----
    If `X` and `Y` have NaN's, only the first contiguous block of non-NaN data
    between both `X` and `Y` is used.

    A binary search is performed to find the indices of `X` that `x`
    lies between.  As such, nonsense results will arise if `X` is not
    sorted along its last dimension.
    """
    y[0] = ppval_1(x, X, Yppc, d)


@nb.njit
def ppinterp_1(f, x, X, Y, d=0):
    """Do a piecewise polynomial interpolation, building the coefficients on the fly."""

    # Find k = index to first valid data site, and
    # K such that K - 1 = index to last valid data site in contiguous range
    #                     of valid data after index k.
    k, K = valid_range_1_two(X, Y)

    # K - 1 == k means there's only one valid data point. Can't interpolate that.
    if K - 1 <= k or np.isnan(x) or x < X[k] or X[K - 1] < x:
        return np.nan

    #   X[k] <= x <= X[k+1]  when  i == k,
    #   X[i] <  x <= X[i+1]  when  k+1 <= i <= K-2.
    i = max(k, np.searchsorted(X[k:K], x) + k - 1)

    Yppc = f(X, Y, i)
    dx = x - X[i]  # >= 0
    return pval(dx, Yppc, d)


@nb.njit
def ppinterp_1_two(f, x, X, Y, Z, d=0):
    """Do two piecewise polynomial interpolations, building the coefficients on the fly."""

    # Find k = index to first valid data site, and
    # K such that K - 1 = index to last valid data site in contiguous range
    #                     of valid data after index k.
    k, K = valid_range_1_two(X, Y)

    # K - 1 == k means there's only one valid data point. Can't interpolate that.
    if K - 1 <= k or np.isnan(x) or x < X[k] or X[K - 1] < x:
        return np.nan, np.nan

    #   X[k] <= x <= X[k+1]  when  i == k,
    #   X[i] <  x <= X[i+1]  when  k+1 <= i <= K-2.
    i = max(k, np.searchsorted(X[k:K], x) + k - 1)

    Yppc = f(X, Y, i)
    Zppc = f(X, Z, i)
    dx = x - X[i]  # >= 0
    return pval(dx, Yppc, d), pval(dx, Zppc, d)


@nb.njit
def ppval_1_two(x, X, Yppc, Zppc, d=0):
    """
    Evaluate two piecewise polynomials.

    As `ppval_1` but a second input, `Zppc`, provides a second set of Piecewise
    Polynomial Coefficients.  Correspondingly, a second output, `z`, is returned.

    `Zppc` must have the same NaN-structure as `Yppc`.
    """

    k, K = valid_range_1_two(X, Yppc[:, -1])

    if K - 1 <= k or np.isnan(x) or x < X[k] or X[K - 1] < x:
        return np.nan, np.nan

    i = max(k, np.searchsorted(X[k:K], x) + k - 1)

    dx = x - X[i]  # >= 0
    return pval(dx, Yppc[i], d), pval(dx, Zppc[i], d)


@nb.njit
def ppval_1_nonan_two(x, X, Yppc, Zppc, d=0):
    """As `ppval_1_two` but knowing there are no NaN's in the inputs X, Yppc, Zppc"""
    if np.isnan(x) or x < X[0] or X[-1] < x:
        return np.nan, np.nan

    i = max(0, np.searchsorted(X, x) - 1)

    dx = x - X[i]  # >= 0
    return pval(dx, Yppc[i], d), pval(dx, Zppc[i], d)


@nb.guvectorize(
    [(nb.f8, nb.f8[:], nb.f8[:, :], nb.f8[:, :], nb.i8, nb.f8[:], nb.f8[:])],
    "(),(n),(n,m),(n,m),()->(),()",
)
def ppval_two(x, X, Yppc, Zppc, d, y, z):
    """
    Evaluate two given Piecewise Polynomials

    The inputs and outputs are as for `ppval`, but another input, `Zppc`, gives
    the second piecewise polynomial, for which a second output `z` is given.

    `Zppc` must have the same NaN-structure as `Yppc`.

    Notes
    -----
    Calling `ppval_two` is faster than calling `ppval` twice, since the former
    calls `np.searchsorted` half as many times.
    """
    y[0], z[0] = ppval_1_two(x, X, Yppc, Zppc, d)



import numpy as np

from .ppinterp import make_pp, ppval_1_two, pval
from .lib import _process_eos

# CHECK VALUE from MATLAB, with densjmd95 (non-Boussinesq) as the eos:
# >> S = linspace(34, 35.5, 20)';
# >> T = linspace(18, 0, 20)';
# >> P = linspace(0, 4000, 20)';
# >> veronis_density(0, S, T, P, 0, 2000) % 1027.098197160422
def veronis_density(
    S,
    T,
    P,
    p1,
    p_ref=0.0,
    p0=None,
    dp=1.0,
    interp="linear",
    eos="gsw",
    grav=None,
    rho_c=None,
):
    """The surface density plus the integrated vertical gradient of Locally
    Referenced Potential Density

    Determines the Veronis density [1]_ [2]_ at vertical position `p1` on a
    cast with hydrographic properties `(S, T, P)`.  The Veronis density is
    the potential density (referenced to `p_ref`) evaluated at `p0` on the
    cast, plus the integral (dP) of the vertical (d/dP) derivative of Locally
    Referenced Potential Density (LRPD) from `P = p0` to `P = p1`.  The
    vertical (d/dP) derivative of LRPD is `rho_S dS/dP + rho_T dT/dP` where
    `rho_S` and `rho_T` are the partial derivatives of density with respect
    to `S` and `T`, and `dS/dP` and `dT/dP` are the derivatives of `S` and
    `T` with respect to `P` in the water column.  If `p0` or `p1` are outside
    the range of `P`, NaN is returned.

    Parameters
    ----------
    S, T, P : 1D ndarray of float

        practical / Absolute salinity, potential / Conservative temperature,
        and pressure or depth of data points on the cast.  `P` must increase
        monotonically along its last dimension.

    p1 : float

        Pressure or depth at which the Veronis density is evaluated

    Returns
    -------
    d : float

        Veronis density

    Other Parameters
    ----------------
    p_ref : float, Default 0.0

        reference pressure or depth for potential density

    p0 : float, Default `P[0]`

        Pressure or depth at which the potential density is evaluated

    dp : float, Default 1.0

        Maximum interval of pressure or depth in trapezoidal numerical
        integration

    interp : str, Default 'linear'

        Method for vertical interpolation.  Use `'linear'` for linear
        interpolation, and `'pchip'` for Piecewise Cubic Hermite Interpolating
        Polynomials.  Other interpolants can be added through the subpackage,
        `ppinterp`.

    eos : function

        Equation of state for the density or specific volume as a function of
        `S`, `T`, and pressure (if non-Boussinesq) or depth (if Boussinesq).

    eos_s_t : function

        Equation of state for the partial derivatives of density or specific
        volume with respect to `S` and `T`.  The inputs are `S`, `T`, and
        pressure (if non-Boussinesq) or depth (if Boussinesq).

    Notes
    -----
    The result of this function can serve as a density label for an
    approximately neutral surface. However, this is NOT the same as a value
    of the Jackett and McDougall (1997) Neutral Density variable. This is
    true even if you were to provide this function with the same cast that
    Jackett and McDougall (1997) used to initially label their Neutral
    Density variable, namely the cast at 188 deg E, 4 deg S, from the Levitus
    (1982) ocean atlas. Some difference would remain, because of differences
    in numerics, and because of a subsequent smoothing step in the Jackett
    and McDougall (1997) algorithm. This function merely allows one to label
    an approximately neutral surface with a density value that is INTERNALLY
    consistent within the dataset where one's surface lives. This function is
    NOT to compare density values against those from any other dataset, such
    as 1997 Neutral Density.

    Examples
    --------
    >>> S = np.linspace(34, 35.5, 20)
    >>> T = np.linspace(18, 0, 20)
    >>> P = np.linspace(0, 4000, 20)
    >>> veronis_density(S, T, P, 2000, eos="jmd95")
    1027.098197160422

    Calculate the Veronis density at 2000 dbar on a water column of linearly
    varying salinity and potential temperature.

    .. [1] Veronis, G. (1972). On properties of seawater defined by temperature,
       salinity, and pressure. Journal of Marine Research, 30(2), 227.

    .. [2] Stanley, McDougall, Barker 2021, Algorithmic improvements to finding
       approximately neutral surfaces, Journal of Advances in Earth System
       Modelling, 13(5).
    """

    # assert(all(size(T) == size(S)), 'T must be same size as S')
    # assert(all(size(P) == size(S)), 'P must be same size as S')
    # assert(isvector(S), 'S, T, P must be 1D. (Veronis density is only useful for one water column at a time!)')
    # assert(isscalar(p0), 'p0 must be a scalar')
    # assert(isscalar(p1), 'p1 must be a scalar')

    if p0 is None:
        p0 = P[0]

    if (
        np.isnan(p0)
        or np.isnan(p1)
        or np.isnan(P[0])
        or p0 < P[0]
        or P[-1] < p0
        or p1 < P[0]
        or P[-1] < p1
    ):
        return np.nan

    eos, eos_s_t = _process_eos(eos, grav, rho_c, need_s_t=True)

    # P[k0-1] <= p0 <= P[k0]  when  k0 == 1
    # P[k0-1] <  p0 <= P[k0]  when  k0 > 1
    # Similarly for k1.
    k0 = max(1, np.searchsorted(P, p0))
    k1 = max(1, np.searchsorted(P, p1))

    ppc_fn = make_pp(interp, kind="1", out="coeffs", nans=True)
    Sppc = ppc_fn(P, S)
    Tppc = ppc_fn(P, T)

    # Integrate from p0 to P[k0]
    d1 = _int_x_k(p0, k0 - 1, dp, P, Sppc, Tppc, eos_s_t)

    # Integrate from P[k0] to P[k1]
    for k in range(k0, k1):
        # Integrate from P[k] to P[k+1], for k=k0 to k=k1-1
        d1 += _int_x_k(P[k], k, dp, P, Sppc, Tppc, eos_s_t)

    # Integrate from p1 to P[k1], and subtract this
    d1 -= _int_x_k(p1, k1 - 1, dp, P, Sppc, Tppc, eos_s_t)

    # Calculate potential density, referenced to p_ref, at p0
    s0, t0 = ppval_1_two(p0, P, Sppc, Tppc)
    d0 = eos(s0, t0, p_ref)

    return d0 + d1


def _int_x_k(p, k, dp, P, Sppc, Tppc, eos_s_t):
    # Integrate from p to P[k+1] using trapezoidal integration with spacing dp or smaller

    # If p == P[k+1], this returns 0.0

    # Number of points between p and P[k], inclusive
    n = int(np.ceil((P[k + 1] - p) / dp)) + 1

    p_ = np.linspace(p, P[k + 1], n)  # intervals are not larger than dp

    # Use piecewise polynomial coefficients as provided. Be sure to pass the
    # index of this part to avoid any issues evaluating a discontinuous piecewise
    # polynomial (probably the derivative, if either) at the knot.
    s_ = np.zeros(n)
    t_ = np.zeros(n)
    dsdp_ = np.zeros(n)
    dtdp_ = np.zeros(n)
    for i in range(n):
        dx = p_[i] - P[k]
        s_[i] = pval(dx, Sppc[k], 0)
        t_[i] = pval(dx, Tppc[k], 0)
        dsdp_[i] = pval(dx, Sppc[k], 1)
        dtdp_[i] = pval(dx, Tppc[k], 1)

    # To use linear interpolation internally, replace the above lines with the following 7 lines
    # dP = P[k+1] - P[k]
    # dsdp_ = (S[k+1] - S[k]) / dP
    # dtdp_ = (T[k+1] - T[k]) / dP
    # s0 = ( S[k] * (P[k+1] - p) + S[k+1] * (p - P[k]) ) / dP
    # t0 = ( T[k] * (P[k+1] - p) + T[k+1] * (p - P[k]) ) / dP
    # s_ = np.linspace(s0, S[k+1], n)
    # t_ = np.linspace(t0, T[k+1], n)

    rs_, rt_ = eos_s_t(s_, t_, p_)
    y_ = rs_ * dsdp_ + rt_ * dtdp_
    return np.trapz(y_, x=p_)


# def veronis_label(p_ref, t_ref, S, T, P, p, pin, eos, eos_s_t, dp=1, interpfn=linear_coeffs):
# 1. Do a temporal neutral_trajectory from (t,i0,j0,p[i0,j0]) to (t_ref,i0,j0,p0)
# 2. Evaluate veronis density at (t_ref, i0, j0, p0)



import numpy as np
import numba as nb

from .lib import diff_1d_samesize
from .ppinterp import ppinterp_1, ppinterp_1_two


def linear_coeffs(X, Y):
    """
    Piecewise Polynomial Coefficients for a linear interpolant

    Parameters
    ----------
    X : ndarray

        Independent variable. Must be same dimensions as `Y`, with any
        dimension possibly a singleton: that is, must be broadcastable
        to the size of `Y`.

        `X` must be monotonically increasing in its last dimension.
        That is, `X[*i,:]` should be monotonically increasing for any
        `i` tuple indexing all but the last dimension.

    Y : ndarray

        Dependent variable.  Must be same dimensions as `X`, with any
        dimension possibly a singleton: that is, must be broadcastable
        to the size of `X`.

    Returns
    -------
    Yppc : ndarray

        Coefficients for a linear interpolant of `Y` to `X`.

    Notes
    -----
    If `X` and `Y` have NaN's, only the first contiguous block of non-NaN data
    between both `X` and `Y` is used.

    Evaluate the piecewise polynomial at `x` as

    >>> y = ppval(x, X, Yppc)

    """
    return _linear_coeffs(X, Y, np.arange(2))


@nb.njit
def linear_coeffs_1(X, Y):
    """
    Coefficients for a single linear interpolant

    Inputs and outputs are as for `linear_coeffs`, but `X` and `Y` both 1D
    arrays of length `n`, and so `Yppc` is a 2D array of size `(n, 2)`.
    """
    C = np.empty((len(Y), 2), dtype=np.float64)
    C[:, 0] = diff_1d_samesize(Y) / diff_1d_samesize(X)
    C[:, 1] = Y
    return C


@nb.guvectorize(
    [(nb.f8[:], nb.f8[:], nb.f8[:], nb.f8[:, :])], "(n),(n),(m)->(n,m)"
)
def _linear_coeffs(X, Y, len2array, C):
    C[:, :] = linear_coeffs_1(X, Y)


@nb.njit
def _linear_coeffs_i(X, Y, i):
    """
    Coefficients of a single Piece of a Linear Interpolant

    Parameters
    ----------
    X : 1D array
        Independent data, monotonically increasing

    Y : 1D array
        Dependent data

    i : int
        Select the interval of `X` that contains the eventual evaluation site
        `x`. Specifically,
            if `i == 0` then `X[0] <= x <= X[1]`, or
            if `1 <= i <= len(X) - 1` then `X[i] < x <= X[i+1]`.

        These facts about `i` are assumed true; they are not checked.

    Returns
    -------
    C1, C0 : float
        Piecewise Polynomial Coefficients that can be evaluated at `x`.

    Notes
    -----
    To interpolate `Y` in terms of `X` at evaluation site `x`, simply evaluate
    the piecewise polynomial whose coefficients are `Yppc` at `x` by
        `pval(x - X[i], (C1, C0), 0)`
    """
    C1 = (Y[i + 1] - Y[i]) / (X[i + 1] - X[i])  # coeff of 1st degree term, x^1
    C0 = Y[i]  # coeff of 0th degree term, x^0

    return C1, C0


@nb.njit
def linear_interp_1(x, X, Y, d=0):
    """Build and evaluate a piecewise polynomial, building only what's needed."""
    return ppinterp_1(_linear_coeffs_i, x, X, Y, d)


@nb.njit
def linear_interp_1_two(x, X, Y, Z, d=0):
    return ppinterp_1_two(_linear_coeffs_i, x, X, Y, Z, d)


@nb.guvectorize(
    [(nb.f8, nb.f8[:], nb.f8[:], nb.i8, nb.f8[:])],
    "(),(n),(n),()->()",
)
def linear_interp(x, X, Y, d, y):
    y[0] = linear_interp_1(x, X, Y, d)


@nb.guvectorize(
    [(nb.f8, nb.f8[:], nb.f8[:], nb.f8[:], nb.i8, nb.f8[:], nb.f8[:])],
    "(),(n),(n),(n),()->(),()",
)
def linear_interp_two(x, X, Y, Z, d, y, z):
    y[0], z[0] = linear_interp_1_two(x, X, Y, Z, d)



"""
Piecewise polynomial interpolation in one dimension.

This package separates 1D interpolation into two steps: 
    1. Compute coefficients for a piecewise polynomial interpolant,
    2. Evaluate the interpolant.

If the interpolant will be evaluated many times, then it is fastest to do Step
1 once, then repeat Step 2 as many times as needed. To do this, build the
coefficients using `*coeffs*` methods, such as `linear_coeffs` and
`pchip_coeffs`. Then evaluate the interpolants using `ppval` methods.

If the interpolant will only be evaluated once or a few times, then it will
be faster to combine Steps 1 and 2, building the polynomial (not a piecewise
polynomial) for just the interval containing the evaluation site, using only as
much data around the evaluation site as is required. To do this, use the
`*interp*` methods, such as `linear_interp` and `pchip_interp`. 

The core numerical methods handle just one interpolation problem; that is, the
evaluation site (`x`) is a scalar and the input independent and dependent
data (`X` and `Y`) are 1D arrays. These functions are named with `"_1"`.

Functions that lack this `"_1"` are "universal", accepting an N-dimensional
array for `x` and (N+1)-dimensional arrays for  `X`, and `Y`, so long as they
are mutually broadcastable from the basic 1D problem, thus allowing multiple
1D interpolation problems to be solved with one function call. Effectively, a
low-level for loop is used to wrap the single 1D problem. Operationally, this
is achieved using `numba`'s `guvectorize` decorator.

If the interpolant must be evaluated many times but the dataset is large and 
memory is limited, then the best strategy is to compute the piecewise 
polynomial coefficients for a single 1D interpolation problem (Step 1), then 
evaluate it (Step 2) as many times as needed. This can be achieved by a
`numba.njit` accelerated for loop over each 1D problem, using the `"_1"` 
function variants. For example, `neutralocean.surface._vertsolve` does this
to solve many individual 1D non-linear root finding problems. 

If performing a single 1D interpolation problem and all input data is finite
(no NaN's), the `"_nonan"` variants provide a small speed advantage. 

If there are two dependent variables that share the same dependent variable,
the `"_two"` variants offer a speed advantage over calling the basic
functions twice, since the `"_two"` variants do the work of locating where
the evaluation site `x` lies within the dependent data `X` only once. This
could easily be extended to handle more dependent variables, of course.

The `make_pp` factory function will help select the desired variant of the
`*coeffs*` and `*interp*` functions.
"""

from .ppinterp import (
    pval,
    ppval_1,
    ppval_1_two,
    ppval,
    ppval_two,
    ppval_1_nonan,
    ppval_1_nonan_two,
)
from .lib import valid_range_1, valid_range_1_two, valid_range
from .tools import make_pp



import numpy as np
import numba as nb

from neutralocean.traj import _ntp_bottle_to_cast
from neutralocean.ppinterp import valid_range_1_two


@nb.njit
def bfs_conncomp1(indptr, indices, root, good):
    """
    Find the Connected Component containing 1 reference location using
    Breadth-first Search

    Parameters
    ----------
    indptr, indices : ndarray of int

        Together, these specify the connectivity of the full graph.  Node `m` is
        adjacent to the nodes `indices[indptr[m] : indptr[m+1]]`.  Note that
        these are simply the `indptr` and `indices` attributes of a the graph
        specified as a sparse matrix in csr_matrix format, i.e. a square matrix
        `G` that has `G[i,j] != 0` whenever nodes `i` and `j` are adjacent.

    root : int

        Index to the chosen root node, where the BFS begins.

    good : ndarray of bool

        A 1D array of logicals for masking out certain nodes in the graph.
        The BFS only operates on "valid nodes" `m` for which `good[m]` is True.
        Note, this function mutates `good`!

    Returns
    -------
    qu : ndarray of int

        The BFS search queue.  A 1D array giving the indices to nodes in the
        connected component containing the `root` node, in the order that the
        BFS discovered them.
        If the root node is valid (`good[root] == True`), then `qu[0] == root`.
        Otherwise, the BFS goes nowhere and `qu` is an empty array.
    """

    N = len(good)
    qu = np.empty(N, dtype=np.int_)

    qt = -1  # Queue Tail
    qh = -1  # Queue Head

    if good[root]:
        # Root node is valid.  Initialize BFS from root node
        qt += 1  # Add root to queue
        qu[qt] = root
        good[root] = False  # mark root as discovered

        while qt > qh:
            qh += 1  # advance head of the queue
            m = qu[qh]  # me node; pop from head of queue

            # loop over neighbour nodes
            for n in indices[indptr[m] : indptr[m + 1]]:
                if good[n]:
                    # n is good and undiscovered
                    qt += 1  # Add n to queue
                    qu[qt] = n
                    good[n] = False  # mark n as discovered
    # else, root node is invalid.  Leave qt as -1, so qu[0:qt+1] is empty

    return qu[0 : qt + 1]


@nb.njit
def bfs_conncomp1_wet(
    indptr, indices, root, s, t, p, S, T, P, tol_p, eos, ppc_fn, p_ml
):
    """
    As in bfs_conncomp1 but extending the perimeter via wetting.

    A breadth-first search begins from the root node `root`, extending through
    wet surface points, i.e. points where `p` is finite.  When an invalid
    node is reached (a dry water column), a neutral tangent plane calculation
    is performed from the surface to this water column.  If successful, and
    if the NTP link reaches this neighbouring cast below the mixed layer,
    then this water column is made "wet": the surface is extended to the
    location on this water column at which NTP intersected, and it is added
    to the BFS so that wetting can proceed as far as possible.

    Parameters
    ----------
    indptr, indices, root :
         As in `bfs_conncomp1`

    s, t, p : ndarray

        2D practical / Absolute salinity and potential / Conservative
        temperature and pressure / depth on the surface.

        Note, this function mutates `s`, `t`, and `p`!

    S, T, P : ndarray

        3D practical / Absolute salinity and potential / Conservative
        temperature and pressure / depth in the ocean

    tol_p : float

        Error tolerance when root-finding to update the pressure or depth of
        the surface in each water column. Units are the same as `P`.

    eos : function

        The equation of state, giving the density or specific volume as a
        function of `S`, `T`, and `P` inputs.

        This should be `@numba.njit` decorated and need not be
        vectorized, as it will be called many times with scalar inputs.

    ppc_fn : function

        Function to compute piecewise polynomial coefficients for an interpolator.
        Construct this as
        `neutralocean.ppinterp.make_pp("linear", kind="1", nans=False)`
        for linear interpolation.  For other interpolants, replace `"linear"`
        (see `make_pp` documentation).

    p_ml : ndarray

        Pressure or depth of the base of the mixed layer.
        If NTP links that enter the mixed layer are to be retained, then pass
        `p_ml = np.full_like(s.shape, -np.inf)`.


    Returns
    -------
    qu : ndarray
        As in `bfs_conncomp1`

    newly_wet : int
        Number of newly wet water columns.

    """

    # N, nk = S.shape
    N = S.shape[0]  # Number of nodes (water columns)

    qu = np.empty(N, dtype=np.int_)

    qt = -1  # Queue Tail
    qh = -1  # Queue Head
    newly_wet = 0

    good = np.isfinite(p)  # Good nodes
    dry = ~good  # nodes to try wetting

    # Flatten lateral dimension of inputs to be 1D.  Use reshape() to get a view.
    # ... removed this!  Should already have been done.
    # good = np.reshape(good, N)
    # dry = np.reshape(dry, N)
    # s = np.reshape(s, N)
    # t = np.reshape(t, N)
    # p = np.reshape(p, N)
    # S = np.reshape(S, (N, nk))
    # T = np.reshape(T, (N, nk))
    # P = np.reshape(P, (N, nk))
    # p_ml = np.reshape(p_ml, -1)  # flatten

    if good[root]:
        # Initialize BFS from root node
        qt += 1  # Add r to queue
        qu[qt] = root
        good[root] = False  # mark r as discovered

        while qt > qh:
            qh += 1  # advance head of the queue
            m = qu[qh]  # me node; pop from head of queue

            # loop over neighbour nodes
            for n in indices[indptr[m] : indptr[m + 1]]:
                if good[n]:
                    # n is good and undiscovered

                    qt += 1  # Add n to queue
                    qu[qt] = n
                    good[n] = False  # mark n as discovered

                elif dry[n]:
                    # n is "dry".  Try wetting.

                    Sn = S[n]
                    Tn = T[n]
                    Pn = P[n]

                    # assume S and T have same nan-structure
                    k, K = valid_range_1_two(Sn, Pn)

                    if K - k <= 1:
                        # At most 1 valid ocean data site.  Can't interpolate
                        # with that.  Mark as not dry, so don't revisit.
                        dry[n] = False
                        continue

                    # Do NTP link from bottle at m to cast at n.
                    s[n], t[n], p[n] = _ntp_bottle_to_cast(
                        s[m], t[m], p[m], Sn, Tn, Pn, k, K, tol_p, eos, ppc_fn
                    )

                    if np.isfinite(p[n]) and p[n] > p_ml[n]:
                        # The NTP connection was successful, and its location
                        # on the neighbouring cast is below the mixed layer.
                        qt += 1  # Add n to queue
                        qu[qt] = n
                        good[n] = False  # mark n as discovered
                        dry[n] = False
                        newly_wet += 1  # augment counter of newly wet casts

    return qu[0 : qt + 1], newly_wet



import numpy as np
import numba as nb


@nb.njit
def diff_1d_samesize(x):
    """
    First difference in one dimension; appends nan so output same size as input.

    A simple function like `np.diff` that numba can work with.
    """

    d = np.empty(x.size, dtype=x.dtype)
    for i in range(d.size):
        d[i] = x[i + 1] - x[i]
    d[-1] = np.nan
    return d


@nb.njit
def valid_range_1(X):
    """The index to the first finite value and to the first NaN after the former

    Parameters
    ----------
    X : 1D array
        Input array possibly containing some NaN elements

    Returns
    -------
    k : int
        First `k` such that `X[k]` is finite.
        If `X[i]` is NaN for all `i`, then `k = len(X)`.

    K : int
        If `X[i]` is NaN for all `i > k`, then `K = len(X)`.
        If `X[i]` is finite for all `i > k`, then `K = len(X)`.
        Otherwise, `K` is the first index after `k` such that `X[K]` is NaN.

    Notes
    -----
    `X[i]` is non-NaN for `i = k, ..., K - 1`.
    `K - k` is the size of the first contiguous block of valid values of `X`.
    """
    k = K = len(X)

    # Find k = index to first valid data site
    for i in range(K):
        if np.isfinite(X[i]):
            k = i
            break

    # Find K, such that K-1 = index to last valid data site
    for i in range(k, K):
        if np.isnan(X[i]):
            K = i
            break

    return k, K


@nb.guvectorize([(nb.f8[:], nb.int64[:], nb.int64[:])], "(n)->(),()")
def valid_range(X, k, K):
    """Indices to the first finite value and to the first NaN after the former,
    along last dimension.

    Parameters
    ----------
    X : ndarray
        Input array possibly containing some NaN elements

    Returns
    -------
    k : ndarray of int
        `k[n]` is the smallest index such that `X[n][k[n]]` is finite.
        If `X[i]` is NaN for all `i`, then `k[n] = len(X[n])`.

    K : ndarray of int
        If `X[n][i]` is NaN for all `i > k[n]`, then `K[n] = len(X[n])`.
        If `X[n][i]` is finite for all `i > k[n]`, then `K[n] = len(X[n])`.
        Otherwise, `K[n]` is the smallest index after `k[n]` such that
        `X[n][K[n]]` is NaN.

    Notes
    -----
    `X[n][i]` is non-NaN for `i = k[n], ..., K[n] - 1`.
    `K[n] - k[n]` is the size of the first contiguous block of valid values of `X[n]`.
    """
    k[0], K[0] = valid_range_1(X)


@nb.njit
def valid_range_1_two(X, Y):
    """Indices bounding the first contiguous block of valid data in two 1D arrays

    Parameters
    ----------
    X, Y : 1D array
        Input arrays of the same length, possibly containing some NaN elements

    Returns
    -------
    k : int
        First `k` such that `X[k]` and `Y[k]` are finite.
        If `X[i]` or `Y[i]` is NaN for all `i`, then `k = len(X)`.

    K : int
        If `X[i]` or `Y[i]` is NaN for all `i > k`, then `K = len(X)`.
        If `X[i]` and `Y[i]` are finite for all `i > k`, then `K = len(X)`.
        Otherwise, `K` is the first index after `k` such that `X[K]` or `Y[K]`
        is NaN.

    Notes
    -----
    `X[i]` and `Y[i]` are both non-NaN for `i = k, ..., K - 1`.
    `K - k` is the size of the first contiguous block of valid pairs of values
    of `(X, Y)`.
    """
    k = K = len(X)

    # Find k = index to first valid data site
    for i in range(K):
        if np.isfinite(X[i]) and np.isfinite(Y[i]):
            k = i
            break

    # Find K, such that K-1 = index to last valid data site
    for i in range(k, K):
        if np.isnan(X[i]) or np.isnan(Y[i]):
            K = i
            break

    return k, K



"""Neutral Trajectory and related functions"""

import numpy as np
import numba as nb

from neutralocean.ppinterp import make_pp, ppval_1_nonan_two
from neutralocean.eos.tools import make_eos
from neutralocean.fzero import guess_to_bounds, brent
from neutralocean.ppinterp import valid_range_1_two
from neutralocean.lib import _process_casts


@nb.njit
def _func(p, sB, tB, pB, Sppc, Tppc, P, eos):
    # Evaluate difference between (a) eos at location on the cast (S, T, P)
    # where the pressure or depth is p, and (b) eos of the bottle (sB, tB, pB)
    # here, eos is always evaluated at the average pressure or depth, (p +
    # pB)/2.
    s, t = ppval_1_nonan_two(p, P, Sppc, Tppc)
    p_avg = (pB + p) * 0.5
    return eos(sB, tB, p_avg) - eos(s, t, p_avg)


def ntp_bottle_to_cast(
    sB,
    tB,
    pB,
    S,
    T,
    P,
    tol_p=1e-4,
    interp="linear",
    eos="gsw",
    grav=None,
    rho_c=None,
):
    """Find the neutral tangent plane from a bottle to a cast

    Finds a point on the cast salinity, temperature, and pressure `(S, T, P)`
    where the salinity, temperature, and pressure `(s, t, p)` is neutrally
    related to a bottle of salinity, temperature and pressure `(sB, tB, pB)`.
    That is, the density of `(s, t, p_avg)` very nearly equals the density
    of `(sB, tB, p_avg)`, where `p_avg = (p + pB) / 2`.  Within `tol_p` of this
    point on the cast, there is a point where these two densities are exactly
    equal.

    Parameters
    ----------
    sB, tB, pB : float

        practical / Absolute salinity, potential / Conservative temperature,
        and pressure or depth of the bottle

    S, T, P : 1D array of float

        practical / Absolute salinity, potential / Conservative temperature,
        and pressure or depth of data points on the cast.  `P` must increase
        monotonically along its last dimension.

    Returns
    -------
    s, t, p : float

        practical / Absolute salinity, potential / Conservative temperature,
        and pressure or depth at a point on the cast that is very nearly
        neutrally related to the bottle.

    Other Parameters
    ----------------
    tol_p : float, Default 1e-4

        Error tolerance in terms of pressure or depth when searching for a root
        of the nonlinear equation.  Units are the same as `P`.

    interp : str, Default 'linear'

        Method for vertical interpolation.  Use `'linear'` for linear
        interpolation, and `'pchip'` for Piecewise Cubic Hermite Interpolating
        Polynomials.  Other interpolants can be added through the subpackage,
        `ppinterp`.

    eos : str or function, Default 'gsw'

        Equation of state for the density or specific volume as a function of
        `S`, `T`, and pressure inputs.  For Boussinesq models, provide `grav`
        and `rho_c`, so this function with third input pressure will be
        converted to a function with third input depth.

        If a function, this should be `@numba.njit` decorated and need not be
        vectorized, as it will be called many times with scalar inputs.

        If a str, can be either `'gsw'` to use TEOS-10
        or `'jmd'` to use Jackett and McDougall (1995) [1]_.

    grav : float, Default None

        Gravitational acceleration [m s-2].  When non-Boussinesq, pass `None`.

    rho_c : float, Default None

        Boussinesq reference density [kg m-3].  When non-Boussinesq, pass `None`.

    Notes
    -----
    .. [1] Jackett and McDougall, 1995, JAOT 12(4), pp. 381-388

    """

    eos = make_eos(eos, grav, rho_c)
    ppc_fn = make_pp(interp, kind="1", out="coeffs", nans=False)

    k, K = valid_range_1_two(S, P)  # S and T have same nan-structure

    return _ntp_bottle_to_cast(sB, tB, pB, S, T, P, k, K, tol_p, eos, ppc_fn)


@nb.njit
def _ntp_bottle_to_cast(sB, tB, pB, S, T, P, k, K, tol_p, eos, ppc_fn):
    """Fast version of `ntp_bottle_to_cast`.

    Parameters
    ----------
    sB, tB, pB, S, T, P, tol_p : float
        See `ntp_bottle_to_cast`

    k, K : int
        `k` is the index to the first finite value in `S + P`.
        `K` is the index to the first NaN value in `S + P` after `k`.
        If `S + P` is all NaN, then `k = len(S)`.
        If `S + P` is all NaN or if `(S + P)[k:]` is all finite, then `K = len(S)`.
        See `neutralocean.ppinterp.lib.valid_range_1`.

    eos : function
        Equation of state for the density or specific volume as a function of
        `S`, `T`, and pressure or depth inputs.

        This function should be `@numba.njit` decorated and need not be
        vectorized, as it will be called many times with scalar inputs.

    ppc_fn : function
        Function that calculates piecewise polynomial coefficients, such as
        returned by `neutralocean.ppinterp.make_pp`

    Returns
    -------
    s, t, p : float
        See ntp_bottle_to_cast
    """

    if K - k > 1:

        # Trim data to valid range and build interpolant
        P = P[k:K]
        Sppc = ppc_fn(P, S[k:K])
        Tppc = ppc_fn(P, T[k:K])

        args = (sB, tB, pB, Sppc, Tppc, P, eos)

        # Search for a sign-change, expanding outward from an initial guess
        lb, ub = guess_to_bounds(_func, pB, P[0], P[-1], args)

        if np.isfinite(lb):
            # A sign change was discovered, so a root exists in the interval.
            # Solve the nonlinear root-finding problem using Brent's method
            p = brent(_func, lb, ub, tol_p, args)

            # Interpolate S and T onto the updated surface
            s, t = ppval_1_nonan_two(p, P, Sppc, Tppc)

        else:
            s, t, p = np.nan, np.nan, np.nan

    else:  # K - k <= 1, so at most one valid data site. Can't interpolate that.
        s, t, p = np.nan, np.nan, np.nan

    return s, t, p


def neutral_trajectory(
    S,
    T,
    P,
    p0,
    vert_dim=-1,
    tol_p=1e-4,
    interp="linear",
    eos="gsw",
    grav=None,
    rho_c=None,
):
    """Calculate a neutral trajectory through a sequence of casts.

    Given a sequence of casts with hydrographic properties `(S, T, P)`, calculate
    a neutral trajectory starting from the first cast at pressure `p0`, or
    starting from a bottle prior to the first cast with hydrographic properties
    `(s0, t0, p0)`.

    Parameters
    ----------
    S, T, P : 2D ndarray or xarray

        1D data specifying the practical / Absolute salinity, and potential /
        Conservative temperature, and pressure / depth down a 1D sequence of casts.
        The first dimension specifies the cast number, while the second provides
        data on that cast; e.g. `S[i, :]` is the salinity down cast `i`.

    p0 : float

        The pressure / depth at which to begin the neutral trajectory on the first cast

    Returns
    -------
    s, t, p : 1D ndarray

        practical / Absolute Salinity, potential / Conservative Temperature,
        and pressure / depth along the neutral trajectory.

    Other Parameters
    ----------------
    vert_dim : int or str, Default -1

        Specifies which dimension of `S`, `T` (and `P` if 3D) is vertical.

        If `S` and `T` are `ndarray`, then `vert_dim` is the `int` indexing
        the vertical dimension of `S` and `T` (e.g. -1 indexes the last
        dimension).

        If `S` and `T` are `xarray.DataArray`, then `vert_dim` is a `str`
        naming the vertical dimension of `S` and `T`.

        Ideally, `vert_dim` is -1.  See `Notes`.

    tol_p : float, Default 1e-4

        Error tolerance when root-finding to update the pressure / depth of
        the surface in each water column. Units are the same as `P`.

    interp : str, Default 'linear'

        Method for vertical interpolation.  Use `'linear'` for linear
        interpolation, and `'pchip'` for Piecewise Cubic Hermite Interpolating
        Polynomials.

    eos : str or function, Default 'gsw'

        Equation of state for the density or specific volume as a function of
        `S`, `T`, and pressure (not depth) inputs.

        If a function, this should be `@numba.njit` decorated and need not be
        vectorized, as it will be called many times with scalar inputs.

        If a str, can be either `'gsw'` to use TEOS-10
        or `'jmd'` to use Jackett and McDougall (1995) [1]_.

    grav : float, Default None

        Gravitational acceleration [m s-2].  When non-Boussinesq, pass `None`.

    rho_c : float, Default None

        Boussinesq reference density [kg m-3].  When non-Boussinesq, pass `None`.

    Notes
    -----
    .. [1] Jackett and McDougall, 1995, JAOT 12(4), pp. 381-388
    """

    eos = make_eos(eos, grav, rho_c)
    ppc_fn = make_pp(interp, kind="1", out="coeffs", nans=False)
    S, T, P = _process_casts(S, T, P, vert_dim)

    nc, nk = S.shape
    # assert(all(size(T) == size(S)), 'T must be same size as S')
    # assert(all(size(P) == size(S)) || all(size(P) == [nk, 1]), 'P must be [nk,nc] or [nk,1]')

    s = np.full(nc, np.nan)
    t = np.full(nc, np.nan)
    p = np.full(nc, np.nan)

    # Loop over casts
    for c in range(0, nc):

        Sc = S[c, :]
        Tc = T[c, :]
        Pc = P[c, :]
        k, K = valid_range_1_two(Sc, Pc)

        if c == 0:
            # Evaluate S and T on first cast at p0
            Sppc = ppc_fn(Pc[k:K], Sc[k:K])
            Tppc = ppc_fn(Pc[k:K], Tc[k:K])
            s[0], t[0] = ppval_1_nonan_two(p0, Pc[k:K], Sppc, Tppc)
            p[0] = p0
        else:

            # Make a neutral connection from previous bottle to the cast (S[c,:], T[c,:], P[c,:])
            s[c], t[c], p[c] = _ntp_bottle_to_cast(
                s[c - 1],
                t[c - 1],
                p[c - 1],
                Sc,
                Tc,
                Pc,
                k,
                K,
                tol_p,
                eos,
                ppc_fn,
            )

        if np.isnan(p[c]):
            # The neutral trajectory incropped or outcropped
            break

    return s, t, p



"""
Density of Sea Water using the Jackett et al. (2006) [1]_ function

Functions:

rho :: computes in-situ density from salinity, potential temperature and
    pressure

rho_s_t :: compute the partial derivatives of in-situ density with
    respect to salinity and potential temperature

rho_p :: compute the partial derivative of in-situ density with
    respect to pressure

Notes:
To make Boussinesq versions of these functions, see 
`neutralocean.eos.tools.make_eos_bsq`.

To make vectorized versions of these functions, see
`neutralocean.eos.tools.vectorize_eos`.

.. [1] Jackett, D. R., McDougall, T. J., Feistel, R., Wright, D. G., &
   Griffies, S. M. (2006). Algorithms for Density, Potential Temperature,
   Conservative Temperature, and the Freezing Temperature of Seawater.
   Journal of Atmospheric and Oceanic Technology, 23(12), 1709–1728.
   https://doi.org/10.1175/JTECH1946.1

History
-------
Code adapted from MOM5.1 (Griffies et al) originally written in Fortran
2022 January 14 - David Hutchinson - Translated into python
2022 January 14 - Geoff Stanley    - code optimization, partial derivatives, check vals
"""

import numpy as np
import numba as nb

# List of coefficients for the rational function
# fmt: off
a0  =  9.9984085444849347e+02
a1  =  7.3471625860981584e+00
a2  = -5.3211231792841769e-02
a3  =  3.6492439109814549e-04
a4  =  2.5880571023991390e+00
a5  = -6.7168282786692355e-03
a6  =  1.9203202055760151e-03
a7  =  1.1798263740430364e-02
a8  =  9.8920219266399117e-08
a9  =  4.6996642771754730e-06
a10 = -2.5862187075154352e-08
a11 = -3.2921414007960662e-12

b0  =  1.0000000000000000e+00 
b1  =  7.2815210113327091e-03
b2  = -4.4787265461983921e-05 
b3  =  3.3851002965802430e-07
b4  =  1.3651202389758572e-10
b5  =  1.7632126669040377e-03
b6  = -8.8066583251206474e-06
b7  = -1.8832689434804897e-10
b8  =  5.7463776745432097e-06
b9  =  1.4716275472242334e-09
b10 =  6.7103246285651894e-06
b11 = -2.4461698007024582e-17
b12 = -9.1534417604289062e-18

epsln = 1.0e-40
# fmt: on


@nb.njit
def rho(s, t, p):
    """
    Parameters
    ----------
    s : float
        Practical salinity [PSS-78]
    t : float
        Potential temperature [ITS-90]
    p : float
        Pressure [dbar]

    Returns
    -------
    rho : float
        In-situ density [kg m-3]
    """

    # Precompute some commonly used terms
    t2 = t * t

    # Rational function for density
    num = (
        a0
        + t * (a1 + t * (a2 + a3 * t))
        + s * (a4 + a5 * t + a6 * s)
        + p * (a7 + a8 * t2 + a9 * s + p * (a10 + a11 * t2))
    )

    inv_den = 1.0 / (
        b0
        + t * (b1 + t * (b2 + t * (b3 + t * b4)))
        + s * (b5 + t * (b6 + b7 * t2) + np.sqrt(s) * (b8 + b9 * t2))
        + p * (b10 + p * t * (b11 * t2 + b12 * p))
        + epsln
    )

    return num * inv_den


@nb.njit
def rho_s_t(s, t, p):
    """
    Parameters
    ----------
    s : float
        Practical salinity [PSS-78]
    t : float
        Potential temperature [ITS-90]
    p : float
        Pressure [dbar]

    Returns
    -------
    rho_s : float
        Partial derivative of in-situ density with respect to salinity [kg m-3 psu-1]

    rho_t : float
        Partial derivative of in-situ density with respect to temperature [kg m-3 degc-1]
    """

    # Precompute some commonly used terms
    t2 = t * t
    sp5 = np.sqrt(s)
    pt = p * t

    # Rational function for density
    num = (
        a0
        + t * (a1 + t * (a2 + a3 * t))
        + s * (a4 + a5 * t + a6 * s)
        + p * (a7 + a8 * t2 + a9 * s + p * (a10 + a11 * t2))
    )

    inv_den = 1.0 / (
        b0
        + t * (b1 + t * (b2 + t * (b3 + t * b4)))
        + s * (b5 + t * (b6 + b7 * t2) + sp5 * (b8 + b9 * t2))
        + p * (b10 + pt * (b11 * t2 + b12 * p))
        + epsln
    )

    # The density is
    #   rho = num / den
    # Taking the partial derivative w.r.t. s gives
    #   rho_s = (num_s - num * den_s / den ) / den
    # and similarly for rho_t

    num_s = a4 + a5 * t + 2.0 * a6 * s + p * a9
    num_t = (
        a1
        + t * (2.0 * a2 + 3.0 * a3 * t)
        + a5 * s
        + 2.0 * a8 * pt
        + 2.0 * a11 * p * pt
    )

    den_s = b5 + t * (b6 + b7 * t2) + sp5 * (1.5 * b8 + 1.5 * b9 * t2)
    den_t = (
        b1
        + t * (2.0 * b2 + t * (3.0 * b3 + 4.0 * b4 * t))
        + s * (b6 + 3.0 * b7 * t2 + 2.0 * b9 * sp5 * t)
        + 3.0 * b11 * pt * pt
        + b12 * p**3
    )

    rho_s = (num_s - num * den_s * inv_den) * inv_den
    rho_t = (num_t - num * den_t * inv_den) * inv_den

    return rho_s, rho_t


@nb.njit(nb.f8(nb.f8, nb.f8, nb.f8))
def rho_p(s, t, p):
    """
    Parameters
    ----------
    s : float
        Practical salinity [PSS-78]
    t : float
        Potential temperature [ITS-90]
    p : float
        Pressure [dbar]

    Returns
    -------
    rho_p : float
        Partial derivative of in-situ density with respect to pressure [kg m-3 dbar-1]
    """

    # Precompute some commonly used terms
    t2 = t * t

    # Rational function for density
    num = (
        a0
        + t * (a1 + t * (a2 + a3 * t))
        + s * (a4 + a5 * t + a6 * s)
        + p * (a7 + a8 * t2 + a9 * s + p * (a10 + a11 * t2))
    )

    inv_den = 1.0 / (
        b0
        + t * (b1 + t * (b2 + t * (b3 + t * b4)))
        + s * (b5 + t * (b6 + b7 * t2) + np.sqrt(s) * (b8 + b9 * t2))
        + p * (b10 + p * t * (b11 * t2 + b12 * p))
        + epsln
    )

    # The density is
    #   rho = num / den
    # Taking the partial derivative w.r.t. p gives
    #   rho_p = (num_p - num * den_p / den ) / den

    num_p = a7 + a8 * t2 + a9 * s + p * (2.0 * a10 + 2.0 * a11 * t2)

    den_p = b10 + p * t * (2.0 * b11 * t2 + 3.0 * b12 * p)

    return (num_p - num * den_p * inv_den) * inv_den



""" Mixed Layer """

from neutralocean.ppinterp import make_pp
from neutralocean.eos.tools import make_eos, vectorize_eos
from neutralocean.lib import _process_casts, _process_vert_dim


def mixed_layer(
    S,
    T,
    P,
    eos="gsw",
    grav=None,
    rho_c=None,
    pot_dens_diff=0.03,
    ref_p=100.0,
    bottle_index=1,
    interp="linear",
    vert_dim=-1,
):
    """Calculate the pressure or depth at the bottom of the mixed layer

    The mixed layer is the pressure or depth at which the potential density
    (referenced to `ref_p`) exceeds the potential density near the surface
    (specifically, the bottle indexed by `bottle_index` on each cast) by an
    amount of `pot_dens_diff`.

    Parameters
    ----------
    S, T : ndarray or xarray.DataArray

        3D practical / Absolute salinity and potential / Conservative
        temperature --- that is, a 2D array of 1D water columns or "casts"

    P : ndarray or xarray.DataArray

        In the non-Boussinesq case, `P` is the 3D pressure, sharing the same
        dimensions as `S` and `T`.

        In the Boussinesq case, `P` is the depth and can be 3D with the same
        structure as `S` and `T`, or can be 1D with as many elements as there
        are in the vertical dimension of `S` and `T`.

        Must increase monotonically along the first dimension (i.e. downwards).

    eos : str or function, Default 'gsw'

        Specification for the equation of state.

        If a str, can be any of the strings accepted by
        `neutralocean.eos.tools.make_eos`,
        e.g. `'jmd95'`, `'jmdfwg06'`, `'gsw'`.

        If a function, must take three inputs corresponding to `S`, `T`, and
        `P`, and output the density (or specific volume).  This form is not
        allowed when `diags` is `True`.  This can be made as, e.g.,
        `eos = neutralocean.eos.make_eos('gsw')`
        for a non-Boussinesq ocean, or as
        `eos = neutralocean.eos.make_eos('gsw', grav, rho_c)`
        for a Boussinesq ocean with `grav` and `rho_c` (see inputs below).

        The function should be `@numba.njit` decorated and need not be vectorized
        -- it will be called many times with scalar inputs.

    grav : float, Default None
        Gravitational acceleration [m s-2].  When non-Boussinesq, pass `None`.

    rho_c : float, Default None
        Boussinesq reference density [kg m-3].  When non-Boussinesq, pass `None`.

    pot_dens_diff : float, Default 0.03

        Difference in potential density [kg m-3] between the near-surface and
        the base of the mixed layer.

    ref_p : float, Default 100.0

        Reference pressure or depth for potential density [dbar or m].

    bottle_index : int, Default 1

        The index for the bottle on each cast where the "near surface" potential
        density is calculated.  Note this index is 0-based.  The Default of 1
        therefore indexes the second bottle in each cast.

    interp : str, Default 'linear'

        Method for vertical interpolation.  Use `'linear'` for linear
        interpolation, and `'pchip'` for Piecewise Cubic Hermite Interpolating
        Polynomials.  Other interpolants can be added through the subpackage,
        `interp1d`.

    vert_dim : int or str, Default -1

        Specifies which dimension of `S`, `T` (and `P` if 3D) is vertical.

        If `S` and `T` are `ndarray`, then `vert_dim` is the `int` indexing
        the vertical dimension of `S` and `T` (e.g. -1 indexes the last
        dimension).

        If `S` and `T` are `xarray.DataArray`, then `vert_dim` is a `str`
        naming the vertical dimension of `S` and `T`.

        Ideally, `vert_dim` is -1.  See `Notes` section of `potential_surf`.


    Returns
    -------
    ML : ndarray

        A 2D array giving the pressure [dbar] or depth [m, positive] of the mixed layer
    """

    # Make the equation of state, if given a str
    eos = make_eos(eos, grav, rho_c)

    # Ensure eos is vectorized. It's okay if eos already was.
    eos = vectorize_eos(eos)

    # Make universal interpolator from the interpolation kernel
    interp_fn = make_pp(interp, kind="u", out="interp")

    # Convert vert_dim from str to int if needed
    vert_dim = _process_vert_dim(vert_dim, S)

    # Convert S, T, P from xarray to numpy arrays if needed, and make casts contiguous in memory
    S, T, P = _process_casts(S, T, P, vert_dim)

    SB = S[..., bottle_index : bottle_index + 1]  # keep singleton trailing dim
    TB = T[..., bottle_index : bottle_index + 1]

    # Calculate potential density difference between each data point and the
    # near-surface bottle
    if eos(34.5, 3, 1000) > 1:  # eos computes in-situ density
        DD = eos(S, T, ref_p) - eos(SB, TB, ref_p)
    else:  # eos computes specific volume
        DD = 1 / eos(S, T, ref_p) - 1 / eos(SB, TB, ref_p)

    # Find the pressure or depth at which the potential density difference
    # exceeds the threshold pot_dens_diff
    return interp_fn(pot_dens_diff, DD, P, 0)



import numpy as np
from neutralocean.eos.tools import make_eos, make_eos_s_t, vectorize_eos
from neutralocean.surface import potential_surf, anomaly_surf, omega_surf
from neutralocean.synthocean import synthocean
from neutralocean.lib import find_first_nan, val_at
from neutralocean.ntp import ntp_epsilon_errors
from neutralocean.grid.rectilinear import build_grid
from neutralocean.grid import divergence

grav = 9.81
rho_c = 1027.5
eos = make_eos("jmd95", grav, rho_c)
eos_s_t = make_eos_s_t("jmd95", grav, rho_c)
eos_ufunc = vectorize_eos(eos)

# Make a simple ocean dataset
ni, nj, nk = 16, 32, 50
wrap = (False, False)  # non-periodic in both horizontal dimensions
S, T, Z, _ = synthocean((ni, nj, nk), wrap=wrap)
# Raise the sea-floor in some casts
# Make one profile be land, and three profiles have a shallower bottom
# (the last having just one valid bottle)
S[1, 1, 45:] = T[1, 1, 45:] = np.nan  # deep ocean
S[2, 1, 5:] = T[2, 1, 5:] = np.nan  # shallow ocean
S[3, 1, 1:] = T[3, 1, 1:] = np.nan  # coastal ocean (1 valid bottle)

# Build grid adjacency and distance information for neutralocean functions
grid = build_grid((ni, nj), wrap)


def test_potential_surf():
    # Test sigma_surf using a prescribed reference depth and isovalue
    z_ref = 0.0
    isoval = 1027.0
    s, t, z, _ = potential_surf(
        S,
        T,
        Z,
        ref=z_ref,
        isoval=isoval,
        eos=eos,
        TOL_P_SOLVER=1e-8,
        diags=False,
    )

    σ = np.ma.masked_invalid(eos_ufunc(s, t, z_ref))
    assert np.ma.allclose(σ, isoval)

    # Calculate surface potential density
    σ_sfc = eos_ufunc(S[:, :, 0], T[:, :, 0], z_ref)

    # Calculate seafloor potential density
    n_good = find_first_nan(S)
    S_bot, T_bot = (val_at(x, n_good - 1) for x in (S, T))
    σ_bot = eos_ufunc(S_bot, T_bot, z_ref)

    σ = eos_ufunc(s, t, z_ref)

    # check for each cast that
    # potential density on surface nearly matches isovalue,
    # or the surface does not intersect this cast (σ is nan) because of one of
    # three conditions: the cast was land, the surface outcropped, or the surface incropped.
    assert np.all(
        (np.abs(σ - isoval) < 1e-8)
        | (np.isnan(σ) & ((n_good == 0) | (isoval < σ_sfc) | (σ_bot < isoval)))
    )


def test_anomaly_surf():
    # Test delta_surf using prescribed reference values and isovalue
    s_ref, t_ref = 34.5, 4.0
    isoval = 0.0
    s, t, z, _ = anomaly_surf(
        S,
        T,
        Z,
        ref=(s_ref, t_ref),
        isoval=isoval,
        eos=eos,
        TOL_P_SOLVER=1e-8,
        diags=False,
    )

    δ = np.ma.masked_invalid(eos_ufunc(s, t, z) - eos_ufunc(s_ref, t_ref, z))
    assert np.ma.allclose(δ, isoval)

    # Calculate surface potential density
    δ_sfc = eos_ufunc(S[:, :, 0], T[:, :, 0], Z[0]) - eos_ufunc(
        s_ref, t_ref, Z[0]
    )

    # Calculate seafloor potential density
    n_good = find_first_nan(S)
    S_bot, T_bot, Z_bot = (val_at(x, n_good - 1) for x in (S, T, Z))
    δ_bot = eos_ufunc(S_bot, T_bot, Z_bot) - eos_ufunc(s_ref, t_ref, Z_bot)

    δ = eos_ufunc(s, t, z) - eos_ufunc(s_ref, t_ref, z)

    # check for each cast that
    # in-situ density anomaly on surface nearly matches isovalue,
    # or the surface does not intersect this cast (δ is nan) because of one of
    # three conditions: the cast was land, the surface outcropped, or the surface incropped.
    assert np.all(
        (np.abs(δ - isoval) < 1e-8)
        | (np.isnan(δ) & ((n_good == 0) | (isoval < δ_sfc) | (δ_bot < isoval)))
    )


def test_omega_surf():
    # Test omega_surf, initialized from a potential density surface
    z0 = 2000.0
    i0, j0 = (int(x / 2) for x in (ni, nj))  # ref cast in middle of domain

    # Calculate initial surface as a potential density surface
    s, t, z, _ = potential_surf(
        S, T, Z, pin_cast=(i0, j0), pin_p=z0, eos=eos, diags=False
    )

    # Calculate divergence of ϵ on initial surface
    e = ntp_epsilon_errors(s, t, z, grid, eos_s_t)
    divg_e_init = divergence(e, grid["edges"])

    # Calculate omega surface, and divergence of ϵ on it
    s, t, z, d = omega_surf(
        S,
        T,
        Z,
        grid,
        pin_cast=(i0, j0),
        p_init=z,
        eos=(eos, eos_s_t),
        diags=True,
    )

    e = ntp_epsilon_errors(s, t, z, grid, eos_s_t)
    divg_e = divergence(e, grid["edges"])

    # Check ratio of ∇⋅ϵ, between initial and converged surface, is small
    rms = lambda x: np.sqrt(np.nanmean(np.square(x)))
    assert rms(divg_e) / rms(divg_e_init) < 1e-4



import numpy as np
import numba as nb

from scipy.sparse import coo_matrix, csr_matrix, find, triu, issparse
from scipy.linalg import issymmetric


@nb.njit
def graph_binary_fcn(edges, nodevals, binary_fcn):
    """
    Apply a binary function to all pairs of data for all pairs of adjacent nodes
    in a graph specified by a sparse COO matrix.

    Parameters
    ----------
    edges : 2D array of int of shape `(2, E)`
    
        The edges in a graph whose nodes are labelled `0, ..., N-1`. 
        Denoting `a = edges[0]` and `b = edges[1]`,
        - the nodes labelled `a[i]` and `b[i]` are adjacent, for `0 <= i <= E-1`
        - must have `0 <= a[i], b[i] <= N-1` for `0 <= i <= E-1`, so `edges` is
        referring to valid nodes. 
        
    nodevals : array-like
        values at the nodes in a graph

    binary_fcn : function
        A `numba.njit` function accepting two numbers and returning one number.

    Notes
    -----
    
    `edges` can instead be a tuple of length 2, with each element
    a tuple of int of length `E`.
    If `edges` is a 2D array of shape `(E, 2)` or is a tuple of length `E`
    with each element a tuple of int of length 2, then one must convert it
    as follows:
        
    >>> edges = numpy.array(edges).transpose()
    
    Returns
    -------
    ev : 1d array
        The value of `binary_fcn` evaluated on each edge.
        That is, `ev[e] = binary_fcn(data[i], data[j])` where `i = row[e]` and
        `j = col[e]`.  If either `i` or `j` is < 0, `ev[e]` is nan.
    """

    a, b = edges

    ev = np.empty(len(a), dtype=nodevals.dtype)
    d = nodevals.reshape(-1)
    for i in range(len(a)):
        ev[i] = binary_fcn(d[a[i]], d[b[i]])
    return ev


def edges_to_graph(edges, N=None, weights=None):
    """
    Convert list of pairs of adjacent nodes into a sparse matrix graph representation

    Parameters
    ----------
    edges : tuple
        Each element is an array of int of length E, where E is the number of
        edges in the grid's graph, i.e. the number of pairs of adjacent water
        columns (including land) in the grid.
        If `edges = (a, b)`, the nodes (water columns) whose linear indices are
        `a[i]` and `b[i]` are adjacent.

    N : int, Default None
        Number of nodes in the graph.
        If None, the number of nodes is assumed to be the largest value in
        `edges`, plus one since elements in `edges` are zero-indexed, i.e. nodes
        are laballed 0, 1, ..., N-1.

    weights : array, Default None
        Weights of the edges in the graph.
        If None, a weight of 1 is given for each edge.

    Returns
    -------
    G : csr_matrix
        Symmetric, sparse, `N` by `N` matrix representation of the undirected
        graph.  `G[m,n] = G[n,m] = weights[i]` where `i` is such that
        `edges[0][i] = m` and `edges[1][i] = n`.
    """
    a, b = edges
    E = len(a)
    if N is None:
        N = np.max((np.max(a), np.max(b))) + 1  # assumes no degree 0 nodes.
    if weights is None:
        weights = np.ones(E, dtype=int)
    G = coo_matrix((weights, edges), shape=(N, N))
    return csr_matrix(G + G.T)


def build_grid(G):
    """
    Convert a dict of (undirected) graphs into a grid dict used by neutralocean.

    Parameters
    ----------
    G : matrix or dict
        If a (sparse) matrix, this specifies the edges in the graph:
            an edge exists between nodes `m` and `n`
            iff `G[m,n]` is nonzero or `G[n,m]` is nonzero.
            If `G` is symmetric, only its upper triangle is used.
            If `G` is not symmetric, it must have only one of `G[m,n]` or
            `G[n,m]` be non-zero; this condition is not checked for.
            The geometric distances are both taken to be 1.0, regardless of the
            data in `G`.
        If a dict, must have `'dist'` and `'distperp'` entries, which are both
            (sparse) matrices with the same sparsity structure.
            The edges in the graph are determined from `G['dist']` as above.
            The distance between adjacent nodes `m` and `n` is given by
            `G['dist'][m,n]` or `G['dist'][n,m]`.
            The distance of the interface between adjacent nodes `m` and `n` is
            given by `G['distperp'][m,n]` or `G['distperp'][n,m]`.

    Returns
    -------
    grid : dict

        grid['edges'] : tuple of length 2
            Each element is an array of int of length E, where E is the number of
            edges in the grid's graph, i.e. the number of pairs of adjacent water
            columns (including land) in the grid.
            If `grid['edges'] = (a, b)`, the nodes (water columns) whose linear
            indices are `a[i]` and `b[i]` are adjacent.

        grid['dist'] : 1d array
            `grid['dist'][i]` is the distance between nodes `a[i]` and `b[i]`,
            where `grid['edges'] = (a, b)`.

        grid['distperp'] : 1d array
            `grid['distperp'][i]` is the distance of the interface between
            nodes `a[i]` and `b[i]`, where `grid['edges'] = (a, b)`.
    """

    grid = dict()
    if isinstance(G, dict):
        if not ("dist" in G and "distperp" in G):
            raise ValueError(
                "If a dict, G must have 'dist' and 'distperp' entries"
            )

        D = triu_if_sym(G["dist"])
        P = triu_if_sym(G["distperp"])

        a, b, dist = find(D)
        A, B, perp = find(P)
        if not np.array_equal(a, A) or not np.array_equal(b, B):
            raise ValueError(
                "Given matrices have different sparsity structure."
            )
        grid["edges"] = (a, b)
        grid["dist"] = dist
        grid["distperp"] = perp

    else:  # G is a matrix
        G = triu_if_sym_structure(G)
        a, b, _ = find(G)

        grid["edges"] = (a, b)
        grid["dist"] = grid["distperp"] = np.ones(len(a))

    return grid


def sym_structure(A):
    """
    Determine whether the sparsity structure of martix `A` is symmetric.
    Returns True iff `A[i,j] == 0` implies `A[j,i] == 0` for all valid.
    Also returns False if A is not a square matrix.
    """

    if len(A.shape) != 2 or A.shape[0] != A.shape[1]:
        return False  # A is not a square matrix

    if issparse(A):
        return ((A != 0) != (A.T != 0)).nnz == 0
    else:
        return np.all((A == 0) == (A.T == 0))


def triu_if_sym(A):
    """Return upper triangle matrix if symmetric"""
    if issymmetric(A):
        return triu(A)
    else:
        return A


def triu_if_sym_structure(A):
    """Return upper triangle matrix if symmetric sparsity structure"""
    if sym_structure(A):
        return triu(A)
    else:
        return A



"""Tools for handling the Equation of State"""
import functools as ft
import numba as nb
import importlib

# List of modules in the same directory as this file, each of which must have
# the following numba.njit'ed functions:  rho, rho_s_t, rho_p.
modules = {"gsw": "specvol", "jmd95": "rho", "jmdfwg06": "rho"}


@ft.lru_cache(maxsize=10)
def _make_eos(eos, derivs, num_p_derivs=0, grav=None, rho_c=None):
    if isinstance(eos, str):
        if eos in modules:
            fcn_name = modules[eos] + derivs
            fn = importlib.import_module(
                "neutralocean.eos." + eos
            ).__getattribute__(fcn_name)
        else:
            raise ValueError(
                f"Equation of state {eos} not (yet) implemented."
                " Currently, eos must be one of " + modules.__str__()
            )

    elif callable(eos):
        fn = eos

    else:
        raise TypeError(
            f"Since eos was not a string, eos must be a callalbe function; found {eos}"
        )

    if grav != None and rho_c != None:
        fn = make_bsq(fn, grav, rho_c, num_p_derivs)

    return fn


def make_eos(eos, grav=None, rho_c=None):
    """Make an equation of state function, possibly modified for Boussinesq

    Parameters
    ----------
    eos : str or function

        If a str, can be `'gsw'` to generate the TEOS-10 specific volume [1]_,
        `'jmd95'` to generate the Jackett and McDougall (1995) in-situ
        density [2]_, or `'jmdfwg06'` to generate the Jackett et al (2006)
        in-situ density [3]_.

        If a function, should be an equation of state as a function of
        practical / Absolute salinity, potential / Conservative temperature,
        and pressure (for non-Boussesinq) / depth (for Boussinesq).

    grav, rho_c : float

        Gravitational acceleration [m s-2] and Boussinesq reference density
        [kg m-3]. If both are provided, the equation of state is modified as
        appropriate for the Boussinesq approximation, in which the third
        argument is depth, not pressure. Specifically, a depth `z` is
        converted to `1e-4 * grav * rho_c * z`, which is the hydrostatic
        pressure [dbar] at depth `z` [m] caused by a water column of density
        `rho_c` under gravity `grav`.

    Returns
    -------
    eos : function

        The desired equation of state.

    .. [1] McDougall, T.J. and P.M. Barker, 2011: Getting started with TEOS-10 and
       the Gibbs Seawater (GSW) Oceanographic Toolbox, 28pp., SCOR/IAPSO WG127,
       SBN 978-0-646-55621-5.

    .. [2] Jackett and McDougall, 1995, JAOT 12(4), pp. 381-388

    .. [3] Jackett, D. R., McDougall, T. J., Feistel, R., Wright, D. G., &
       Griffies, S. M. (2006). Algorithms for Density, Potential Temperature,
       Conservative Temperature, and the Freezing Temperature of Seawater.
       Journal of Atmospheric and Oceanic Technology, 23(12), 1709–1728.
       https://doi.org/10.1175/JTECH1946.1
    """

    return _make_eos(eos, "", 0, grav, rho_c)


def make_eos_s_t(eos, grav=None, rho_c=None):
    """Make a function for the partial S and T derivatives of an equation of state

    Parameters
    ----------
    eos, grav, rho_c :
        See `make_eos`

    Returns
    -------
    eos_s_t : function

        Function returning two outputs, namely the partial derivatives with
        respect to its first two arguments (practical / Absolute salinity and
        potential / Conservative temperature) of the desired equation of
        state.
    """

    return _make_eos(eos, "_s_t", 0, grav, rho_c)


def make_eos_p(eos, grav=None, rho_c=None):
    """Make a function for the partial P derivative of an equation of state

    Parameters
    ----------
    eos, grav, rho_c :
        See `make_eos`

    Returns
    -------
    eos_p : function

        Function returning the partial derivative with respect to the third
        argument (pressure) of the desired equation of state.
    """

    return _make_eos(eos, "_p", 1, grav, rho_c)


@ft.lru_cache(maxsize=10)
def make_bsq(fn, grav, rho_c, num_p_derivs=0):
    """Make a Boussinesq version of a given equation of state (or its partial derivative(s))

    Parameters
    ----------
    fn : function
        Function with (salinity, temperature, pressure) as inputs.  Typically
        this is the equation of state, returning the density or specific volume.
        However, it can also be a function for partial derivative(s) of the
        equation of state with respect to salinity, temperature, or pressure.

    grav : float
        Gravitational acceleration [m s-2]

    rho_c : float
        Boussinesq reference density [kg m-3]

    num_p_derivs : int, Default 0
        Number of `p` partial derivatives that relate `fn` to the equation of
        state.  For example,

        - if `fn` is the equation of state, or its partial derivative (of
          any order, with respect to salinity or temperature, pass 0.
        - if `fn` is the partial derivative of the equation of state with
          respect to pressure, pass 1.
        - if `fn` is the second partial derivative of the equation of state
          with respect to salinity and pressure (i.e. ∂²ρ/∂S∂p), pass 1.

    Returns
    -------
    fn_bsq : function
        Boussinesq version of `fn`.
        The inputs to `fn_bsq` are (salinity, temperature, depth).
    """

    # Hydrostatic conversion from depth [m] to pressure [dbar]
    z_to_p = 1e-4 * grav * rho_c

    if num_p_derivs == 0:
        # Slight optimization for later: don't multiply by factor when factor == 1
        @nb.njit
        def fn_bsq(s, t, z):
            return fn(s, t, z * z_to_p)

    else:
        factor = z_to_p**num_p_derivs

        @nb.njit
        def fn_bsq(s, t, z):
            return fn(s, t, z * z_to_p) * factor

    return fn_bsq


@ft.lru_cache(maxsize=10)
def vectorize_eos(eos):
    """Convert an `eos` function that takes scalar inputs into one taking arrays.

    Parameters
    ----------
    eos : function
        Any function taking three scalar inputs and returning one scalar output,
        such as the equation of state.  Note this does not work for functions
        returning multiple outputs (i.e. a tuple), such as a function returning
        the partial derivatives of the equation of state.

    Returns
    -------
    eos_vec : function
        A `@numba.vectorize`'d version of `eos`, which can take array inputs and
        returns one array output.  The array inputs' shape need not match
        exactly, but must be broadcastable to each other.
    """

    @nb.vectorize
    def eos_vec(s, t, p):
        return eos(s, t, p)

    return eos_vec



import numpy as np
import pytest

from neutralocean.eos.tools import vectorize_eos

from neutralocean.eos.jmd95 import rho as rho_jmd95
from neutralocean.eos.jmd95 import rho_s_t as rho_s_t_jmd95
from neutralocean.eos.jmd95 import rho_p as rho_p_jmd95

from neutralocean.eos.jmdfwg06 import rho as rho_jmdfwg06
from neutralocean.eos.jmdfwg06 import rho_s_t as rho_s_t_jmdfwg06
from neutralocean.eos.jmdfwg06 import rho_p as rho_p_jmdfwg06

from neutralocean.eos.gswc import rho as rho_gsw
from neutralocean.eos.gswc import rho_s_t as rho_s_t_gsw
from neutralocean.eos.gswc import rho_p as rho_p_gsw

from neutralocean.eos.gsw import specvol as specvol_gsw
from neutralocean.eos.gsw import specvol_s_t as specvol_s_t_gsw
from neutralocean.eos.gsw import specvol_p as specvol_p_gsw

checkval_jmd95 = (35.5, 3.0, 3000.0, 1041.83267)
rho_jmd95_ufunc = vectorize_eos(rho_jmd95)

# Check Values from Jackett and McDougall (1995) Appendix, p. 388
# and Jackett et al (2006) Appendix A, p. 1723
@pytest.mark.parametrize(
    "eos,checkval,decimals",
    [
        (rho_jmd95, checkval_jmd95, 5),
        (rho_jmdfwg06, (35.0, 25.0, 2000.0, 1031.65056056576), 11),
        (rho_jmdfwg06, (20.0, 20.0, 1000.0, 1017.72886801964), 11),
        (rho_jmdfwg06, (40.0, 12.0, 8000.0, 1062.95279820631), 11),
        (rho_gsw, (35.0, 25.0, 2000.0, 1031.534727150376), 12),
        (specvol_gsw, (35.0, 25.0, 2000.0, 9.694293111803510e-04), 18),
    ],
)
def test_checkval(eos, checkval, decimals):
    assert np.round(eos(*checkval[:-1]), decimals=decimals) == checkval[-1]


def test_jmd95_ufunc_scalar():
    res = rho_jmd95_ufunc(*checkval_jmd95[:-1])
    assert np.round(res, decimals=5) == checkval_jmd95[-1]


def test_jmd95_ufunc_array():
    # Smoketest: broadcasting
    s = np.ones((4, 5), dtype=float) * checkval_jmd95[0]
    t = np.ones((5,), dtype=float) * checkval_jmd95[1]
    p = checkval_jmd95[2]
    res = rho_jmd95_ufunc(s, t, p)
    assert res.shape == s.shape
    assert np.all(np.round(res, decimals=5) == checkval_jmd95[-1])


@pytest.mark.parametrize(
    "eos,eos_s_t,eos_p",
    [
        (rho_jmd95, rho_s_t_jmd95, rho_p_jmd95),
        (rho_jmdfwg06, rho_s_t_jmdfwg06, rho_p_jmdfwg06),
        (rho_gsw, rho_s_t_gsw, rho_p_gsw),
        (specvol_gsw, specvol_s_t_gsw, specvol_p_gsw),
    ],
)
def test_rho_derivs(eos, eos_s_t, eos_p):
    """Check first partial derivatives by centred differences

    Parameters
    ----------
    eos : function
        Function of (S, T, P) for the Equation of State.
        Can give density or specific volume.

    eos_s_t : function
        Function of (S, T, P) returning a tuple of length two, giving the
        partial derivatives of `eos` with respect to `S` and `T`.

    eos_p : function
        Function of (S, T, P) returning the partial derivatives of `eos` with
        respect to `P`.

    Returns
    -------
    None.

    Raises
    ------
    AssertionError
        If the results of `eos_s_t` and `eos_p` at a (hardcoded) checkvalue
        disagree considerably with an approximation of partial derivatives
        calculated by evaluating `eos` using centred finite differences.

    """

    s, t, p = (35.0, 25.0, 2000.0)
    ds, dt, dp = (1e-4, 1e-4, 1e-1)

    rs_centred = (eos(s + ds, t, p) - eos(s - ds, t, p)) / (2.0 * ds)
    rt_centred = (eos(s, t + dt, p) - eos(s, t - dt, p)) / (2.0 * dt)
    rp_centred = (eos(s, t, p + dp) - eos(s, t, p - dp)) / (2.0 * dp)

    rs, rt = eos_s_t(s, t, p)
    rp = eos_p(s, t, p)

    assert np.isclose(rs_centred, rs, atol=0, rtol=1e-8)
    assert np.isclose(rt_centred, rt, atol=0, rtol=1e-8)
    assert np.isclose(rp_centred, rp, atol=0, rtol=1e-8)



import functools as ft
from importlib import import_module


@ft.lru_cache(maxsize=10)
def make_pp(
    interpolant="linear",
    kind="u",
    out="coeffs",
    nans=True,
    num_dep_vars=1,
):
    """Make function for piecewise polynomial interpolation."""

    if interpolant not in ("linear", "pchip"):
        raise ValueError(
            f"Expected `interpolant` in ('lienar', 'pchip'); got {interpolant}"
        )

    if kind not in ("1", "u"):
        raise ValueError(f"Expected `kind` in ('1', 'u'); got {kind}")

    if out not in ("coeffs", "interp"):
        raise ValueError(f"Expected `out` in ('coeffs', 'interp'); got {out}")

    if num_dep_vars not in (1, 2):
        raise ValueError(
            "Expected `num_dep_vars` in (1, 2); got {num_dep_vars}"
        )

    if out == "coeffs" and num_dep_vars != 1:
        raise ValueError(
            "With `out='coeffs'`, currently only handles `num_dep_vars=1`"
        )

    if nans == False:
        if out == "interp" or kind == "u":
            raise ValueError(
                "With `nans=False`, expected `out='coeffs'` and `kind='1'`."
            )

    # Begin programmatically generating the function name to import
    if kind == "1":
        kind = "_1"
    elif kind == "u":
        kind = ""

    if nans == False and interpolant == "pchip":
        nans = "_nonan"
    else:
        # if interpolant == "linear", just use regular coeffs_1 function.
        nans = ""

    if num_dep_vars == 1:
        num_dep_vars = ""
    elif num_dep_vars == 2:
        num_dep_vars = "_two"

    fcn_name = interpolant + "_" + out + kind + nans + num_dep_vars

    # Below is equivalent to
    # from neutralocean.ppinterp.`interpolant` import `fcn_name`
    return import_module(
        "neutralocean.ppinterp." + interpolant
    ).__getattribute__(fcn_name)



import numpy as np


def build_grid(dims, periodic, dxC=1.0, dyC=1.0, dyG=1.0, dxG=1.0):
    """
    Build pairs of adjacent grid points and associated geometry for a rectilinear grid

    Parameters
    ----------
    dims : tuple of int
        Dimensions of the grid.
        `dims[n]` is the number of grid cells in the n'th direction, for n = 1,2.

    periodic : tuple of bool
        Specifies periodicity.
        The n'th dimension is periodic when `periodic[n]` is True.

    dxC : float or ndarray, Default 1.0
        Distance between adjacent grid points in the 1st ('x') dimension.
        Specifically, `dxC[i,j]` is the distance between the centers of cells
        `(i,j)` and `(i-1,j)`.

    dyC : float or ndarray, Default 1.0
        Distance between adjacent grid points in the 2nd ('y') dimension.
        Specifically, `dyC[i,j]` is the distance between the centers of cells
        `(i,j)` and `(i,j-1)`.

    dyG : float or ndarray, Default 1.0
        Distance (in the 2nd, 'y' dimension) of the face between grid points
        that are adjacent in the 1st ('x') dimension.
        Lives at same location as `dxC`.
        Specifically, `dyG[i,j]` is the distance of the interface between
        cells `(i,j)` and `(i-1,j)`.

    dxG : float or ndarray, Default 1.0
        Distance (in the 1st, 'x' dimension) of the face between grid points
        that are adjacent in the 2nd ('y') dimension.
        Lives at same location as `dyC`.
        Specifically, `dxG[i,j]` is the distance of the interface between
        cells `(i,j)` and `(i,j-1)`.

    Notes
    -----
    The above notes for `dxC`, `dyC`, `dyG`, `dxG` are valid for
    `0 <= i <= ni-1` and `0 <= j <= nj-1`, where `dims == (ni, nj)`,
    using Python's -1 indexing notation:

    - `dxC[0,j]` is the distance between the centers of cells `(0,j)` and `(ni-1,j)`
    - `dyG[0,j]` is the distance  of  the  face between cells `(0,j)` and `(ni-1,j)`
    - `dyC[0,j]` is the distance between the centers of cells `(i,0)` and `(i,nj-1)`
    - `dxG[0,j]` is the distance  of  the  face between cells `(i,0)` and `(i,nj-1)`

    If `periodic[0]` is False, `dxC[0,j]` and `dyG[0,j]` are not used.

    If `periodic[1]` is False, `dyC[i,0]` and `dxG[i,0]` are not used.

    Broadcasting: The shape of `dxC`, `dyC`, `dyG`, `dxG` will be broadcast to
    shape `dims == (ni, nj)`.

    Example on a latitude-longitude grid: the meridional grid distances are
    constant, so `dyC` and `dyG` can be passed as scalars rather than 2D arrays;
    likewise, the zonal grid distances only depend on latitude, so `dxC` and
    `dxG` can be passed as 1D arrays of length `nj`.

    Returns
    -------
    grid : dict
        Containing the following:

        edges : tuple of length 2
            Each element is an array of int of length `E`, where `E` is the number of
            edges in the grid's graph, i.e. the number of pairs of adjacent water
            columns (including land) in the grid.
            If `edges = (a, b)`, the nodes (water columns) whose linear indices are
            `a[i]` and `b[i]` are adjacent.

        dist : 1d array
            Horizontal distance between adjacent water columns (nodes).
            `dist[i]` is the distance between nodes whose linear indices are
            `edges[0][i]` and `edges[1][i]`.

        distperp : 1d array
            Horizontal distance of the face between adjacent water columns (nodes).
            `distperp[i]` is the distance of the interface between nodes whose
            linear indices are `edges[0][i]` and `edges[1][i]`.
    """
    grid = dict()
    grid["edges"] = _build_edges(dims, periodic)
    grid["dist"] = _build_edgedata(dims, periodic, (dxC, dyC))
    grid["distperp"] = _build_edgedata(dims, periodic, (dyG, dxG))
    return grid


def _build_edges(dims, periodic):
    """
    Build list of pairs of adjacent grid points, numbered 0, ..., N-1, on a
    rectilinear grid
    """

    ndim = len(dims)  # Number of dimensions in the grid
    assert ndim == 2, "currently only works for 2D grids."
    assert (
        len(periodic) == ndim
    ), "periodic must be a tuple the same length as dims."

    ni, nj = dims
    N = ni * nj  # number of nodes

    i1 = int(not periodic[0])  # 0 when periodic, 1 when non-periodic
    j1 = int(not periodic[1])
    i2 = ni - i1  # ni when periodic, ni - 1 when non-periodic
    j2 = nj - j1

    Ei = i2 * nj  # number of edges in i dimension
    Ej = ni * j2  # number of edges in j dimension
    E = Ei + Ej  #  number of edges in total

    a, b = (np.empty(E, dtype=int) for x in (0, 1))  # prealloc space

    # Build linear indices to the entire grid
    idx = np.arange(N).reshape((ni, nj))

    # Handle first dimension
    if periodic[0]:
        a[:Ei] = idx.reshape(-1)
        b[:Ei] = np.roll(idx, 1, axis=0).reshape(-1)
    else:
        a[:Ei] = idx[i1:, :].reshape(-1)
        b[:Ei] = idx[:i2, :].reshape(-1)

    # Handle second dimension
    if periodic[1]:
        a[Ei:] = idx.reshape(-1)
        b[Ei:] = np.roll(idx, 1, axis=1).reshape(-1)
    else:
        a[Ei:] = idx[:, j1:].reshape(-1)
        b[Ei:] = idx[:, :j2].reshape(-1)

    return a, b


def _build_edgedata(dims, periodic, data):
    """
    Build a 1D array of `data` in the same order as constructed by `_build_edges`

    Parameters
    ----------
    dims, periodic :
        See `build_grid`

    data : tuple of ndarray
        Data that lives on edges, i.e. between nodes (water columns).

        With `dims == (ni, nj)`:

        - `data[0][i, j]` lives between nodes [i, j] and [i-1, j], for `1 <= i <= ni-1`.

        - `data[0][0, j]` lives between nodes [0, j] and [ni-1, j], if
        `periodic[0]` is True, and is not used if `periodic[0]` is False.

        - `data[1][i, j]` lives between nodes [i, j] and [i, j-1], for `1 <= j <= nj-1`.

        - `data[1][i, 0]` lives between nodes [i, 0] and [i, nj-1], if
        `periodic[1]` is True, and is not used if `periodic[1]` is False.

        Each element of `data` will be broadcast to the size `dims`.

        Example: `data = (xdist, ydist)` where `xdist` is the distance
        between water columns in the x (first) dimension, and `ydist` is the
        distance between water columns in the y (second) dimension.

    Returns
    -------
    edgedata : array
        1D array of `data`.
        If `a, b = _build_edges(dims, periodic)`, then `edgedata[i]` is the
        value from `data` that corresponds to the interface between the grid
        cells indexed by `a[i]` and `b[i]`.

    """
    i1 = int(not periodic[0])  # 0 when periodic, 1 when non-periodic
    j1 = int(not periodic[1])
    x = np.broadcast_to(data[0], dims)
    y = np.broadcast_to(data[1], dims)
    return np.concatenate((x[i1:, :].reshape(-1), y[:, j1:].reshape(-1)))


def edgedata_to_maps(edgedata, dims, periodic):
    """
    Convert 1D array of data living on edges into two 2D arrays, one for each
    spatial dimension, for a rectilinear grid

    Parameters
    ----------
    edgedata : array
        1D array of data that lives on edges of the grid's graph, given in the
        same order as the edges cosntructed from `_build_edges`.
        If `a, b = _build_edges(dims, periodic)`, then `edgedata[i]` is the
        data that lives at the interface between the grid cells indexed by
        `a[i]` and `b[i]`.

    dims, periodic :
        See `build_grid`

    Returns
    -------

    Fi, Fj : ndarray
        2D arrays of shape `dims`.  `Fi` contains the values from `edgedata`
        that correspond to data living between grid points in the first
        dimension, and similarly for `Fj` but in the second dimension.
    """

    ni, nj = dims

    i1 = int(not periodic[0])  # 0 when periodic, 1 when non-periodic
    j1 = int(not periodic[1])

    Ei = (ni - i1) * nj  # number of edges in i dimension

    Fi, Fj = (np.full((ni, nj), np.nan) for x in (0, 1))

    Fi[i1:, :] = edgedata[:Ei].reshape((ni - i1, nj))
    Fj[:, j1:] = edgedata[Ei:].reshape((ni, nj - j1))

    return Fi, Fj



"""
Density of Sea Water using the Jackett and McDougall 1995 [1]_ function

Functions:

rho :: computes in-situ density from salinity, potential temperature and
    pressure

rho_s_t :: compute the partial derivatives of in-situ density with
    respect to salinity and potential temperature

rho_p :: compute the partial derivative of in-situ density with
    respect to pressure

Notes:
To make Boussinesq versions of these functions, see 
`neutralocean.eos.tools.make_eos_bsq`.

To make vectorized versions of these functions, see
`neutralocean.eos.tools.vectorize_eos`.

.. [1] Jackett and McDougall, 1995, JAOT 12[4], pp. 381-388
"""

import numpy as np
import numba as nb


# fmt: off
# Coefficients nonlinear equation of state in pressure coordinates for
# 1. density of fresh water at p = 0
eosJMDCFw0 =    999.842594  
eosJMDCFw1 =    6.793952e-02
eosJMDCFw2 = -  9.095290e-03
eosJMDCFw3 =    1.001685e-04
eosJMDCFw4 = -  1.120083e-06
eosJMDCFw5 =    6.536332e-09
# 2. density of sea water at p = 0
eosJMDCSw0 =    8.244930e-01
eosJMDCSw1 = -  4.089900e-03
eosJMDCSw2 =    7.643800e-05
eosJMDCSw3 = -  8.246700e-07
eosJMDCSw4 =    5.387500e-09
eosJMDCSw5 = -  5.724660e-03
eosJMDCSw6 =    1.022700e-04
eosJMDCSw7 = -  1.654600e-06
eosJMDCSw8 =    4.831400e-04
# coefficients in pressure coordinates for
# 3. secant bulk modulus K of fresh water at p = 0
eosJMDCKFw0 =   1.965933e+05 # .== original * 10
eosJMDCKFw1 =   1.444304e+03 # .== original * 10
eosJMDCKFw2 = - 1.706103e+01 # .== original * 10
eosJMDCKFw3 =   9.648704e-02 # .== original * 10
eosJMDCKFw4 = - 4.190253e-04 # .== original * 10
# 4. secant bulk modulus K of sea water at p = 0
eosJMDCKSw0 =   5.284855e+02 # .== original * 10
eosJMDCKSw1 = - 3.101089e+00 # .== original * 10
eosJMDCKSw2 =   6.283263e-02 # .== original * 10
eosJMDCKSw3 = - 5.084188e-04 # .== original * 10
eosJMDCKSw4 =   3.886640e+00 # .== original * 10
eosJMDCKSw5 =   9.085835e-02 # .== original * 10
eosJMDCKSw6 = - 4.619924e-03 # .== original * 10
# 5. secant bulk modulus K of sea water at p
eosJMDCKP0 =   3.186519e+00
eosJMDCKP1 =   2.212276e-02
eosJMDCKP2 = - 2.984642e-04
eosJMDCKP3 =   1.956415e-06
eosJMDCKP4 =   6.704388e-03
eosJMDCKP5 = - 1.847318e-04
eosJMDCKP6 =   2.059331e-07
eosJMDCKP7 =   1.480266e-04
eosJMDCKP8 =   2.102898e-05 # .== original / 10
eosJMDCKP9 = - 1.202016e-06 # .== original / 10
eosJMDCKP10 =   1.394680e-08 # .== original / 10
eosJMDCKP11 = - 2.040237e-07 # .== original / 10
eosJMDCKP12 =   6.128773e-09 # .== original / 10
eosJMDCKP13 =   6.207323e-11 # .== original / 10
# The above coeffs need to be defined in the function for
# compatability with @numba.njit
# fmt: on

# If ndarray inputs are needed, it is best to use @nb.vectorize.  That is,
# apply `.tools.vectorize_eos`.  A vectorized function specified
# for scalars is about twice as fast as a signatureless njit'ed function
# applied to ndarrays.
@nb.njit
def rho(s, t, p):
    """Fast JMD95 [1]_ in-situ density.

    Parameters
    ----------
    s : float
        Practical salinity [PSS-78]
    t : float
        Potential temperature [IPTS-68]
    p : float
        Pressure [dbar]

    Returns
    -------
    rho : float
        JMD95 in-situ density [kg m-3]

    Notes
    -----
    This function is derived from `densjmd95.m`, documented below. Input
    checks and expansion of variables have been removed. That code used
    arrays to store coefficients, and also began by multiplying the
    input pressure by 10 (dbar -> bar). The coefficients, modified to
    not need this multiplication by 10, have been taken out of arrays
    and simply hardcoded into the later expressions.  The polynomial
    calculations have also been optimized to favour faster, nested
    multiplications.

    As such, the output of this function differs from the output of the
    original `densjmd95.m` function, though the difference is at the
    level of machine precision.

    .. highlight:: matlab
    .. code-block:: matlab

        % DENSJMD95    Density of sea water
        %=========================================================================
        %
        % USAGE:  dens = densjmd95(S,Theta,P)
        %
        % DESCRIPTION:
        %    Density of Sea Water using Jackett and McDougall 1995 (JAOT 12)
        %    polynomial (modified UNESCO polynomial).
        %
        % INPUT:  (all must have same dimensions)
        %   S     = salinity    [psu      (PSS-78)]
        %   Theta = potential temperature [degree C (IPTS-68)]
        %   P     = pressure    [dbar]
        %       (P may have dims 1x1, mx1, 1xn or mxn for S(mxn) )
        %
        % OUTPUT:
        %   dens = density  [kg/m^3]
        %
        % AUTHOR:  Martin Losch 2002-08-09  (mlosch@mit.edu)


        % Jackett and McDougall, 1995, JAOT 12(4), pp. 381-388

        % created by mlosch on 2002-08-09
        % $Header: /u/gcmpack/MITgcm/utils/matlab/densjmd95.m,v 1.2 2007/02/17 23:49:43 jmc Exp $
        % $Name:  $

    .. [1] Jackett and McDougall, 1995, JAOT 12[4], pp. 381-388
    """

    s1o2 = np.sqrt(s)

    # fmt: off
    # The secant bulk modulus
    K = (           1.965933e+05 + t*( 1.444304e+03 + t*(-1.706103e+01 + t*(9.648704e-02 + t*-4.190253e-04)))
        +   s *    (5.284855e+02 + t*(-3.101089e+00 + t*( 6.283263e-02 + t*-5.084188e-04))
        +   s1o2 * (3.886640e+00 + t*( 9.085835e-02 + t* -4.619924e-03) ))
        + p * (     3.186519e+00 + t*( 2.212276e-02 + t*(-2.984642e-04 + t* 1.956415e-06))
        +   s *    (6.704388e-03 + t*(-1.847318e-04 + t*  2.059331e-07) + s1o2*1.480266e-04)
        + p * (     2.102898e-05 + t*(-1.202016e-06 + t*  1.394680e-08)
        +   s *   (-2.040237e-07 + t*( 6.128773e-09 + t*  6.207323e-11)) )) )

    # The in-situ density
    rho = (
                   999.842594   + t*( 6.793952e-02 + t*(-9.095290e-03 + t*( 1.001685e-04 + t*(-1.120083e-06 + t*6.536332e-09))))
        + s * (    8.244930e-01 + t*(-4.089900e-03 + t*( 7.643800e-05 + t*(-8.246700e-07 + t*  5.387500e-09)))
        + s1o2 * (-5.724660e-03 + t*( 1.022700e-04 + t* -1.654600e-06))
        + s    *   4.831400e-04
        )) / (1.0 - p / K )
    # fmt: on
    return rho


# If we specified a signature for scalar inputs and outputs, such as
#   @nb.njit(nb.typeof((1.0, 1.0))(nb.f8, nb.f8, nb.f8))
# then we would not be able to make a function that uses @numba.vectorize to
# wrap this function.  The tuple output messes that up.  However, we could
# make a function that uses @numba.guvectorize to wrap this function, e.g.
#   @nb.guvectorize([(nb.f8, nb.f8, nb.f8, nb.f8[:], nb.f8[:])], "(),(),()->(),()")
#   def eos_vec(s, t, p, eos_s, eos_t):
#       eos_s[0], eos_t[0] = rho_s_t(s, t, p)
# However, this appears to be about two times slower than just using @numba.njit
# without a signature specification, and calling the rho_s_t function with ndarray
# objects.  So, we'll just use @nb.njit with no signature.
@nb.njit
def rho_s_t(s, t, p):
    """
    Fast salinity and potential temperature partial derivatives of JMD95 in-situ density

    Parameters
    ----------
    s, t, p : float
        See `rho`

    Returns
    -------
        rho_s : float
            Partial derivative of JMD95 in-situ density with respect to
            practical salinity `s` [kg m-3 psu-1]
        rho_t : float
            Partial derivative of JMD95 in-situ density with respect to
            potential temperature `t`   [kg m-3 degC-1]

    Notes
    -----
    This function is derived from `rho`.
    """

    # INPUT CHECKS REMOVED

    s1o2 = np.sqrt(s)

    # fmt: off
    # The secant bulk modulus
    K =    (        eosJMDCKFw0 + t*(eosJMDCKFw1 + t*(eosJMDCKFw2 + t*(eosJMDCKFw3 + t*eosJMDCKFw4)))
        +   s *    (eosJMDCKSw0 + t*(eosJMDCKSw1 + t*(eosJMDCKSw2 + t* eosJMDCKSw3))
        +   s1o2 * (eosJMDCKSw4 + t*(eosJMDCKSw5 + t* eosJMDCKSw6) ))
        + p * (     eosJMDCKP0 + t*(eosJMDCKP1 + t*(eosJMDCKP2 + t*eosJMDCKP3))
        +   s *    (eosJMDCKP4 + t*(eosJMDCKP5 + t* eosJMDCKP6) + s1o2*eosJMDCKP7)
        + p * (     eosJMDCKP8 + t*(eosJMDCKP9 + t* eosJMDCKP10)
        +   s *    (eosJMDCKP11 + t*(eosJMDCKP12 + t* eosJMDCKP13)) )) )


    # The partial derivative of K with respect to s
    # K_s =      ( eosJMDCKSw0 + t*(eosJMDCKSw1 + t*(eosJMDCKSw2 + t* eosJMDCKSw3))
    #     + s1o2 * (1.5*eosJMDCKSw4 + t*(1.5*eosJMDCKSw5 + t* (1.5*eosJMDCKSw6)) )
    #     + p * ( eosJMDCKP4 + t*(eosJMDCKP5 + t* eosJMDCKP6) + s1o2*(1.5*eosJMDCKP7)
    #     + p * ( eosJMDCKP11 + t*(eosJMDCKP12 + t* (eosJMDCKP13)) )) )

    # The partial derivative of K with respect to t
    # K_t =    (     (eosJMDCKFw1 + t*(2.0*eosJMDCKFw2 + t*(3.0*eosJMDCKFw3 + t*(4.0*eosJMDCKFw4))))
    #     +   s *    (eosJMDCKSw1 + t*(2.0*eosJMDCKSw2 + t*(3.0*eosJMDCKSw3))
    #     +   s1o2 * (eosJMDCKSw5 + t*(2.0*eosJMDCKSw6) ))
    #     + p * (     eosJMDCKP1 + t*(2.0*eosJMDCKP2 + t*(3.0*eosJMDCKP3))
    #     +   s *    (eosJMDCKP5 + t*(2.0*eosJMDCKP6))
    #     + p * (     eosJMDCKP9 + t*(2.0*eosJMDCKP10)
    #     +   s *    (eosJMDCKP12 + t*(2.0*eosJMDCKP13)) )) )

    # work =    (    eosJMDCFw0 + t*(eosJMDCFw1 + t*(eosJMDCFw2 + t*(eosJMDCFw3 + t*(eosJMDCFw4 + t*eosJMDCFw5))))
    #     + s *    ( eosJMDCSw0 + t*(eosJMDCSw1 + t*(eosJMDCSw2 + t*(eosJMDCSw3 + t*eosJMDCSw4)))
    #     + s1o2 * ( eosJMDCSw5 + t*(eosJMDCSw6 + t*eosJMDCSw7))
    #     + s    *   eosJMDCSw8   # from here up is the density of sea water at p = 0
    #     )) * p / (K - p) # this prepares for the final rho_s and rho_t computations.

    # The partial derivative of sea water at p = 0, with respect to s
    # rho_0_S =    (     eosJMDCSw0 + t*(    eosJMDCSw1 + t*(    eosJMDCSw2 + t*(eosJMDCSw3 + t*eosJMDCSw4)))
    #     + s1o2 * ( 1.5*eosJMDCSw5 + t*(1.5*eosJMDCSw6 + t*(1.5*eosJMDCSw7)))
    #     + s    * ( 2.0*eosJMDCSw8))

    # The partial derivative of sea water at p = 0, with respect to t
    # rho_0_T =    ( eosJMDCFw1 + t*(2.0*eosJMDCFw2 + t*(3.0*eosJMDCFw3 + t*(4.0*eosJMDCFw4 + t*(5.0*eosJMDCFw5))))
    #     + s *    ( eosJMDCSw1 + t*(2.0*eosJMDCSw2 + t*(3.0*eosJMDCSw3 + t*(4.0*eosJMDCSw4)))
    #     + s1o2 * ( eosJMDCSw6 + t*(2.0*eosJMDCSw7))))

    # The in-situ density is defined as
    #  rho = rho_0 / (1 - p / K)
    # Taking the partial derivative w.r.t S gives
    #           /            rho_0 * p * K_S   \       1
    #  rho_s = | rho_0_S -  _________________   |  _________
    #           \            (1 - p/K) * K^2   /   1 - p / K
    # This is re-written as
    #           /                rho_0 * p * K_S    \       1
    #  rho_s = | rho_0_S * K -  __________________   |  _________
    #           \                     (K - p)       /     K - p
    # Similarly for rho_t.

    # rho_s = (rho_0_S * K - work * K_s) / (K - p)
    # rho_t = (rho_0_T * K - work * K_t) / (K - p)

    # The following expressions are faster and require less memory, though appear more gruesome
    rho_s = (
                 (     eosJMDCSw0 + t*(    eosJMDCSw1 + t*(    eosJMDCSw2 + t*(eosJMDCSw3 + t*eosJMDCSw4)))
        + s1o2 * ( 1.5*eosJMDCSw5 + t*(1.5*eosJMDCSw6 + t*(1.5*eosJMDCSw7)))
        + s    * ( 2.0*eosJMDCSw8)) # rho_0_S
        * K
        - 
              (    eosJMDCFw0 + t*(eosJMDCFw1 + t*(eosJMDCFw2 + t*(eosJMDCFw3 + t*(eosJMDCFw4 + t*eosJMDCFw5))))
        + s *    ( eosJMDCSw0 + t*(eosJMDCSw1 + t*(eosJMDCSw2 + t*(eosJMDCSw3 + t*eosJMDCSw4)))
        + s1o2 * ( eosJMDCSw5 + t*(eosJMDCSw6 + t*eosJMDCSw7))
        + s    *   eosJMDCSw8   # from here up is the density of sea water at p = 0
        )) # rho_0
        * p / (K - p) 
        *
              ( eosJMDCKSw0 + t*(eosJMDCKSw1 + t*(eosJMDCKSw2 + t* eosJMDCKSw3))
        + s1o2 * (1.5*eosJMDCKSw4 + t*(1.5*eosJMDCKSw5 + t* (1.5*eosJMDCKSw6)) )
        + p * ( eosJMDCKP4 + t*(eosJMDCKP5 + t* eosJMDCKP6) + s1o2*(1.5*eosJMDCKP7)
        + p * ( eosJMDCKP11 + t*(eosJMDCKP12 + t* (eosJMDCKP13)) )) )  # K_s
        ) / (K - p)
    
    rho_t = (
                 ( eosJMDCFw1 + t*(2.0*eosJMDCFw2 + t*(3.0*eosJMDCFw3 + t*(4.0*eosJMDCFw4 + t*(5.0*eosJMDCFw5))))
        + s *    ( eosJMDCSw1 + t*(2.0*eosJMDCSw2 + t*(3.0*eosJMDCSw3 + t*(4.0*eosJMDCSw4)))
        + s1o2 * ( eosJMDCSw6 + t*(2.0*eosJMDCSw7)))) # rho_0_T
        * K
        - 
              (    eosJMDCFw0 + t*(eosJMDCFw1 + t*(eosJMDCFw2 + t*(eosJMDCFw3 + t*(eosJMDCFw4 + t*eosJMDCFw5))))
        + s *    ( eosJMDCSw0 + t*(eosJMDCSw1 + t*(eosJMDCSw2 + t*(eosJMDCSw3 + t*eosJMDCSw4)))
        + s1o2 * ( eosJMDCSw5 + t*(eosJMDCSw6 + t*eosJMDCSw7))
        + s    *   eosJMDCSw8   # from here up is the density of sea water at p = 0
        )) # rho_0
        * p / (K - p) 
        * 
             (     (eosJMDCKFw1 + t*(2.0*eosJMDCKFw2 + t*(3.0*eosJMDCKFw3 + t*(4.0*eosJMDCKFw4))))
        +   s *    (eosJMDCKSw1 + t*(2.0*eosJMDCKSw2 + t*(3.0*eosJMDCKSw3))
        +   s1o2 * (eosJMDCKSw5 + t*(2.0*eosJMDCKSw6) ))
        + p * (     eosJMDCKP1 + t*(2.0*eosJMDCKP2 + t*(3.0*eosJMDCKP3))
        +   s *    (eosJMDCKP5 + t*(2.0*eosJMDCKP6))
        + p * (     eosJMDCKP9 + t*(2.0*eosJMDCKP10)
        +   s *    (eosJMDCKP12 + t*(2.0*eosJMDCKP13)) )) ) # K_t
        ) / (K - p)

    # fmt: on

    return rho_s, rho_t


@nb.njit
def rho_p(s, t, p):
    """
    Fast pressure derivative of JMD95 in-situ density.

    Parameters
    ----------
    s, t, p : float
        See `rho`

    Returns
    -------
    rho_p : float
        Partial derivative of JMD95 in-situ density with respect to
        pressure `p` [kg m-3 dbar-1]

    Notes
    -----
    This function is derived from `rho`.
    """

    s1o2 = np.sqrt(s)

    # fmt: off
    K2 = (
        p * ( eosJMDCKP8 + t*(eosJMDCKP9 + t*eosJMDCKP10)  
        + s *  (eosJMDCKP11 + t*(eosJMDCKP12 + t*eosJMDCKP13)) ) 
        )

    K1plusK2 = (
        K2 +
            eosJMDCKP0 + t*(eosJMDCKP1 + t*(eosJMDCKP2 + t*eosJMDCKP3))            # \__ these 2 lines are K1
            + s * ( eosJMDCKP4 + t*(eosJMDCKP5 + t*eosJMDCKP6) + eosJMDCKP7*s1o2 ) # /  
        )
    # K == bulkmod
    K = (   eosJMDCKFw0 + t*(eosJMDCKFw1 + t*(eosJMDCKFw2 + t*(eosJMDCKFw3 + t*eosJMDCKFw4)))  # secant bulk modulus of fresh water at the surface
        + s * (   eosJMDCKSw0 + t*(eosJMDCKSw1 + t*(eosJMDCKSw2 + t*eosJMDCKSw3)) 
            + s1o2 * (eosJMDCKSw4 + t*(eosJMDCKSw5 + t*eosJMDCKSw6) ) 
        )  # secant bulk modulus of sea water at the surface
        + p * K1plusK2 # secant bulk modulus of sea water at pressure z
        )

    rho = (
        eosJMDCFw0 + t*(eosJMDCFw1 + t*(eosJMDCFw2 + t*(eosJMDCFw3 + t*(eosJMDCFw4 + t*eosJMDCFw5)))) # density of freshwater at the surface
        + s * ( eosJMDCSw0 + t*(eosJMDCSw1 + t*(eosJMDCSw2 + t*(eosJMDCSw3 + t*eosJMDCSw4)))
            + s1o2 * ( eosJMDCSw5 + t*(eosJMDCSw6 + t*eosJMDCSw7) )
            + s * eosJMDCSw8
        ) # density of sea water at the surface
        )
    
    # (K1 + 2 * K2) is K_p, the partial derivative of K w.r.t p
    rho_p = rho * (K - p * (K1plusK2 + K2)) / (K - p)**2
    # fmt: on

    return rho_p



"""Functions to handle updating the pressure / depth of a surface, by 
solving a nonlinear equation in the vertical dimension of each water column"""
import numpy as np
import numba as nb
import functools

from neutralocean.fzero import guess_to_bounds, brent
from neutralocean.ppinterp import ppval_1_nonan_two, valid_range_1_two


@functools.lru_cache(maxsize=10)
def _make_vertsolve(eos, ppc_fn, ans_type):

    if ans_type == "omega":

        @nb.njit
        def f(*args):
            _vertsolve_omega(*args, eos, ppc_fn)
            return None

    elif ans_type == "potential":

        @nb.njit
        def f(*args):
            return _vertsolve(*args, eos, ppc_fn, _zero_potential)

    elif ans_type == "anomaly":

        @nb.njit
        def f(*args):
            return _vertsolve(*args, eos, ppc_fn, _zero_anomaly)

    else:
        raise NameError(f'Unknown ans_type "{ans_type}"')

    return f


@nb.njit
def _vertsolve(S, T, P, ref, d0, tol_p, eos, ppc_fn, zero_func):

    shape2D = S.shape[0:-1]
    s = np.full(shape2D, np.nan)
    t = np.full(shape2D, np.nan)
    p = np.full(shape2D, np.nan)

    for n in np.ndindex(shape2D):

        # Select this water column
        Sn = S[n]
        Tn = T[n]
        Pn = P[n]

        k, K = valid_range_1_two(Sn, Pn)  # S and T have same nan-structure

        if K - k > 1:

            Pn = Pn[k:K]
            Sppcn = ppc_fn(Pn, Sn[k:K])
            Tppcn = ppc_fn(Pn, Tn[k:K])

            args = (Sppcn, Tppcn, Pn, ref, d0, eos)

            # Use mid-pressure as initial guess
            pn = (Pn[0] + Pn[-1]) * 0.5

            # Search for a sign-change, expanding outward from an initial guess
            lb, ub = guess_to_bounds(zero_func, pn, Pn[0], Pn[-1], args)

            if np.isfinite(lb):
                # A sign change was discovered, so a root exists in the interval.
                # Solve the nonlinear root-finding problem using Brent's method
                p[n] = brent(zero_func, lb, ub, tol_p, args)

                # Interpolate S and T onto the updated surface
                s[n], t[n] = ppval_1_nonan_two(p[n], Pn, Sppcn, Tppcn, 0)

    return s, t, p


@nb.njit
def _vertsolve_omega(s, t, p, S, T, P, ϕ, tol_p, eos, ppc_fn):
    # Note!  mutates s, t, p

    for n in np.ndindex(s.shape):
        ϕn = ϕ[n]
        if ϕn == 0.0:
            continue  # leave s,t,p unchanged (avoid creating errors of size tol_p)

        # Select this water column
        Sn = S[n]
        Tn = T[n]
        Pn = P[n]

        k, K = valid_range_1_two(Sn, Pn)  # S and T have same nan-structure
        if K - k > 1 and np.isfinite(ϕn):

            pn = p[n]
            Pn = Pn[k:K]

            # Build interpolant's coefficients for this water column
            Sppcn = ppc_fn(Pn, Sn[k:K])
            Tppcn = ppc_fn(Pn, Tn[k:K])

            # Evaluate difference between
            # (a) eos at location on the cast where the pressure or depth is p, and
            # (b) eos at location on the cast where the pressure or depth is pn
            # (where the surface currently is) plus the density perturbation d.
            # Part (b) is precomputed.  Here, eos always evaluated at the
            # pressure or depth of the original position, pn, i.e. we calculate
            # locally referenced potential density.
            args = (Sppcn, Tppcn, Pn, pn, eos(s[n], t[n], pn) + ϕn, eos)

            # Search for a sign-change, expanding outward from an initial guess
            lb, ub = guess_to_bounds(_zero_potential, pn, Pn[0], Pn[-1], args)

            if np.isfinite(lb):
                # A sign change was discovered, so a root exists in the interval.
                # Solve the nonlinear root-finding problem using Brent's method
                p[n] = brent(_zero_potential, lb, ub, tol_p, args)

                # Interpolate S and T onto the updated surface
                s[n], t[n] = ppval_1_nonan_two(p[n], Pn, Sppcn, Tppcn)

            else:
                # Ensure s,t,p all have the same nan structure
                s[n], t[n], p[n] = np.nan, np.nan, np.nan

        else:
            # ϕ is nan, or only one grid cell so cannot interpolate.
            # Ensure s,t,p all have the same nan structure
            s[n], t[n], p[n] = np.nan, np.nan, np.nan

    return None


@nb.njit
def _zero_potential(p, Sppc, Tppc, P, ref_p, isoval, eos):
    # Evaluate the potential density in a given cast, minus a given isovalue
    s, t = ppval_1_nonan_two(p, P, Sppc, Tppc, 0)
    return eos(s, t, ref_p) - isoval


@nb.njit
def _zero_anomaly(p, Sppc, Tppc, P, ref, isoval, eos):
    # Evaluate the specific volume (or in-situ density) anomaly in a given cast,
    # minus a given isovalue
    s, t = ppval_1_nonan_two(p, P, Sppc, Tppc)
    return eos(s, t, p) - eos(ref[0], ref[1], p) - isoval
